import os
from neointerface import NeoInterface
from typing import List
import pandas as pd
import numpy as np
from logger.logger import logger


class ModelManager(NeoInterface):
    """
    Python class to manage metadata nodes (such as nodes with label Class, Relationship, Term) along with relationships between them
    as well as to return information about them
    """
    URI_MAP = {
        "Class": {"properties": ["label"]},
        "Property": {"properties": ["Class.label", "label"]},
        "Relationship": {"properties": ["relationship_type"],
                         "neighbours": [
                             {"label": "Class", "relationship": "FROM", "property": "label"},
                             {"label": "Class", "relationship": "TO", "property": "label"},
                         ]},
        "Term": {"properties": ["Codelist Code", "Term Code"]},
        "Method": {"properties": ["parent_id", "id"]},
    }
    SCD = 50  # SUBCLASS_OF allowed Depth
    RDFSLABEL = "rdfs:label"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.verbose:
            logger.info(f"---------------- {self.__class__} initialized -------------------")

    def gen_default_reltype(self, to_label: str) -> str:
        """
        Default relationship type name is generated from the label of the :TO Class
        """
        return f'{to_label}'

    def gen_default_reltypes_list(self, rels: list) -> [{}]:
        """
        Updates list of dicts with relationships {'from':...,'to':...,'type':...}
        replacing 'type':None with default relationship type
        """
        return [{**rel, **{'type': (rel['type'] if rel['type'] else self.gen_default_reltype(rel['to']))}}
                for rel in rels
                ]

    def create_class(self, classes, merge=True, merge_on: List[str]=None) -> [list]:
        """
        :param classes:  List of string labels or a list of property dictionaries to give to the new class(es)
                         created. For example:
                         classes = ['class1', 'class2' ...] OR classes = [{"label": 'class1'}, {"label": 'class2'] ...]
                         Will both result in the creation of two new classes with labels class1 and class2 respectively.
        :param merge:    boolean - if True use MERGE statement to create nodes to avoid duplicate classes
                             TODO: address question "would we want to ever allow multiple classes with the same name??"
        :param merge_on: Optional list of property names MERGED on when classes is a list of property dictionaries. This
                         can be used to selectively rename certain properties on an existing node rather than
                         creating a new one, For example:

                            With classes = [{'label': 'class1', 'type': 'new_type'}] and merge on = ['label']
                            If a class node with 'label' = 'class1' already exists with 'type' = 'old_type' rather
                            than creating a new node 'class1' will be updated with 'type' = 'new_type'

                         When not merge_on is not set the default behaviour is to merge on all properties.
        :return:         A list of lists that contain a single dictionary with keys 'label', 'neo4j_id' and 'neo4j_labels'
                         EXAMPLE: [ [{'label': 'A', 'neo4j_id': 0, 'neo4j_labels': ['Class']}],
                                    [{'label': 'B', 'neo4j_id': 1, 'neo4j_labels': ['Class']}]
                                  ]
        """

        # Maintain backwards compatibility:
        if type(classes) == str:
            classes = [classes]

        assert type(classes) == list, "Classes must be a list"
        assert type(merge) == bool, "Merge must be a bool"
        if merge_on:
            assert merge, "Merge_on requires merge = true"

        if type(classes[0]) == str:
            # Note class_item is the result of unwinding $classes, which in this case is a list of string labels.
            apoc_action = f"""
                CALL apoc.{'merge' if merge else 'create'}.node(
                    ['Class'], {{label: class_item}}{', {}, {}' if merge else ''}
                ) YIELD node
            """
        elif type(classes[0]) == dict:
            # Note class_item is the result of unwinding $classes, which in this case is a list of property
            # dictionaries. So a new node would be created/merged with that given dictionary.
            if not merge_on:
                apoc_action = f"""
                    CALL apoc.{'merge' if merge else 'create'}.node(
                        ['Class'], class_item{', {}, {}' if merge else ''}
                    ) YIELD node
                """
            else:
                ident_props = f'{{`{merge_on[0]}`: class_item["{merge_on[0]}"]'
                for prop in merge_on[1:]:
                    ident_props += f', `{prop}`: class_item["{prop}"]'
                ident_props += '}'

                apoc_action = f"""
                    CALL apoc.merge.node(
                        ['Class'], {ident_props}, class_item, class_item
                    ) YIELD node
                """
                # Example resulting apoc_action format(merge):
                # With ident props = ['label'] the resulting action would be:
                # CALL apoc.merge.node(
                #   ['Class'], {`label`: class_item[`label`]}, class_item, class_item
                # ) YIELD node
                # Which matches on 'label then sets the properties to class_item
        else:
            raise AssertionError('Classes must be a list of strings or dict')

        q = f"""
        WITH $classes as classes 
        UNWIND classes as class_item 
        {apoc_action}
        RETURN node as class
        ORDER by node
        """

        params = {'classes': classes}

        if self.verbose:
            logger.debug(f"""
            query: {q}
            parameters: {params}
            """)

        return self.query(q, params, return_type='neo4j.Result')

    def delete_class(self, values: list, identifier='label'):
        """
        Deletes a class and any associated relationships or controlled terminology
        :param values: list of class labels or property values (if identifier is changed) to match classes for removal
        :param identifier: Class property to use in combination with values for identification.
        :return:
        """
        # TODO: address the issue of invalid Methods linked to the deleted entities.
        q = f"""
        MATCH (class:Class)
        WHERE class[$identifier] in $values
        OPTIONAL MATCH (class)-[:HAS_CONTROLLED_TERM]->(term:Term)
        OPTIONAL MATCH (class)-[:TO|FROM]-(rel:Relationship)
        DETACH DELETE class, term, rel
        """
        params = {'values': values, 'identifier': identifier}

        return self.query(q, params, return_type='neo4j.Result')

    def get_missing_classes(self, values: list, identifier='label'):
        """
        Samples a list of class property values to determine if a set of classes are missing from the database.
        :param values: List of property values eg: [class1, class2]
        :param identifier: Property name to be used when identifying classes eg: 'label'
        :return: A list of any values not found in neo4j.
        """

        q = f"""
        MATCH (c:Class)
        WHERE c.{identifier} in $values
        RETURN collect(c.{identifier}) as existing
        """
        params = {'values': values}
        existing_classes = self.query(q, params)[0].get('existing')

        missing_classes = set(values) - set(existing_classes)
        return missing_classes

    def set_short_label(self, label: str, short_label: str) -> None:
        "One the class with :Class{label:{label}} - sets property 'short_label value to the provided"
        q = """
        MATCH (c:Class)
        WHERE c.label = $label
        SET c.short_label = $short_label
        """
        params = {'label': label, 'short_label': short_label}
        self.query(q, params)

    def create_related_classes_from_list(self, rel_list: [[str, str, str]], identifier='label') -> [str]:
        """
        Create `Class` and `Relationship` nodes between them, as specified by rel_list

        Given a list of relationship triplets, perform 2 operations:
        1)  Identify all the unique names, and create new nodes labeled "Class",
            each new node has one of the unique names stored in an attribute named "label"
        2)  Adds <-[:FROM]-(:Relationship{relationship_type:'...'})-[:TO]-> relationships to pairs of the newly-created nodes,
            as specified by the triplets in the elements of rel_list

        EXAMPLE:  if rel_list is  [  ["Study", "Site", "Site],  ["Study", "Subject", "Subject]  ]
                  then 3 new `Class`-labeled nodes will be created, with "label" attributes respectively
                  valued "Study", "Site" and "Subject",
                  plus <-[:FROM]-(:Relationship{relationship_type:'Site'})-[:TO]-> relationship from "Study" to "Site",
                  and <-[:FROM]-(:Relationship{relationship_type:'Subject'})-[:TO]-> relationship from "Study" to "Subject"

        :param rel_list: A list of 3-element lists, indicating a relationship among nodes of type `Class`
                         EXAMPLE:   [
                                        ["Study", "Site", "Site'],
                                        ["Study", "Subject", "Subject"],
                                        ["Subject", "Race", "Race"]
                                    ]
        :param identifier: String class property used to identify to & from classes
        :return: List of all the class names; repeated ones are taken out
        """

        # Identify all the unique class names in inner elements of rel_list
        class_set = set()  # Empty set
        for rel in rel_list:
            # [:2] in order to get only the first two items (classes) out of rel_list; [2] is the relationship type
            class_set = class_set.union(rel[:2])  # The Set Union operation will avoid duplicates

        class_list = sorted(list(class_set))  # Convert the final set back to list

        self.create_relationship(rel_list, identifier, match_classes=False)

        return class_list

    def create_relationship(self, rel_list: List[List[str]], identifier='label', match_classes=True) -> [str]:
        """
        Create relationship nodes between two specified classes as defined in rel_list.
        For example:
        1.  With rel_list = [ ['class1', 'class2', 'example'] ]
            A new relationship node will be created between nodes with "label", as specified by the
            identifier, 'class1' and 'class2' with a relationship_type property = 'example'.
            This relationship node also includes 'FROM.Class.label' and 'TO.Class.label' properties
            regardless of the class identifier.

        2.  With rel_list = [ ['class1', 'class2', 'example','false'] ]
            A new relationship node will be created between nodes with "label", as specified by the
            identifier, 'class1' and 'class2' with a relationship_type property = 'example' and optional_relationship = 'false'.
            This relationship node also includes 'FROM.Class.label' and 'TO.Class.label' properties
            regardless of the class identifier.
        Note if no relationship type is included in example-1, a default is generated via gen_default_reltype() 
        and if no relationship type is included in example-2, optional_relationship would be considered as relationship type 
        and optional relationship would be ignored
        :param rel_list: A list of relationships represented as lists
        :param identifier: String class property used to identify to & from classes
        :param match_classes: Boolean, If false classes are merged rather than matched which will create them
                              if they do not already exist.
        :return: A list of created relationships = rel_list if all relationships were created successfully
        """

        q = f"""
        UNWIND $rels as rel
        CALL apoc.do.when(size(rel)<=3,
        "WITH rel[0] as from_identity, rel[1] as to_identity, rel[2] as rel_type
        {'MATCH' if match_classes else 'MERGE'} (from:Class {{`{identifier}`:from_identity}})
        {'MATCH' if match_classes else 'MERGE'} (to:Class {{`{identifier}`:to_identity}})   
        MERGE (from)<-[:FROM]-(rel_node:Relationship{{relationship_type:rel_type}})-[:TO]->(to)
        SET rel_node.`FROM.Class.label` = from.label
        SET rel_node.`TO.Class.label` = to.label
        RETURN collect([from.`{identifier}`, to.`{identifier}`, rel_node.relationship_type]) as rels", 
        "WITH rel[0] as from_identity, rel[1] as to_identity, rel[2] as rel_type, rel[3] as optional
        {'MATCH' if match_classes else 'MERGE'} (from:Class {{`{identifier}`:from_identity}})
        {'MATCH' if match_classes else 'MERGE'} (to:Class {{`{identifier}`:to_identity}})   
        MERGE (from)<-[:FROM]-(rel_node:Relationship{{relationship_type:rel_type, relationship_optional:optional}})-[:TO]->(to)
        SET rel_node.`FROM.Class.label` = from.label
        SET rel_node.`TO.Class.label` = to.label
        RETURN collect([from.`{identifier}`, to.`{identifier}`, rel_node.relationship_type, rel_node.relationship_optional]) as rels
        ",
        {{rel:rel}})
        YIELD value
        RETURN collect(value.rels) as rels
        """

        res = self.query(q, {
            "rels": [(r if len(r)>=3 else r + [self.gen_default_reltype(to_label=r[1])]) for r in rel_list]
        })

        if res:
            result = [np.array(i).flatten().tolist() for i in res[0].get('rels')]
            return result
        else:
            return []

    def delete_relationship(self, rel_list: [[str, str, str]], identifier='label'):
        """
        Deletes specified relationships between classes.
        :param rel_list: List of relationships to be deleted in the following format:
                         [from class prop value, to class prop value, relationship type]
                         For example: With identifier = 'label' and
                         rel_list = [['class1', 'class2', 'Example'], ...]
                         Relationships of type "Example" between classes with labels = 'class1' and 'class2'
                         would be deleted.
        :param identifier: String class property to be used when identifying classes.
        :return:
        """

        q = f"""
        UNWIND $rels as rel
        WITH rel[0] as from, rel[1] as to, rel[2] as type
        MATCH (:Class{{`{identifier}`:from}})<-[:FROM]-(rel:Relationship {{relationship_type:type}})-[:TO]->(:Class{{`{identifier}`:to}})
        DETACH DELETE rel
        """
        params = {"rels": rel_list}
        return self.query(q, params, return_type='neo4j.Result')

    def get_all_classes(self) -> [str]:
        ""
        return [c['Class'] for c in self.get_all_classes_with_nodeids()]

    def get_all_classes_with_nodeids(self, include_id=False, sort=True) -> [dict]:
        """
        Get all the existing Class names, optionally including their Neo4j ID, and optionally sorted
        :param include_id:  If True, also include the Neo4j ID's
        :param sort:        If True, sort the results by name
        :return:            A list of dictionaries, with keys "Class" (for the name) and "_id_Class"
                            EXAMPLE, with include_id=False:
                                [{'Class': 'car'}, {'Class': 'boat'}]
                            EXAMPLE, with include_id=True:
                                [{'Class': 'car', "_id_Class": 88}, {'Class': 'boat', "_id_Class": 91}]
        """
        q = '''
            MATCH (class:Class)
            RETURN class.label as Class, class.short_label as short_label
            '''

        if include_id:
            q += " , id(class) as _id_Class "

        if sort:
            q += " ORDER BY class.label"

        return self.query(q)

    def get_all_classes_props(self, props: [str]) -> [dict]:
        """
        Retrieve a list of property values for all classes.
        :param props: List of properties ro retrieve ie ['label', ...]
        :return: A list of dictionaries, with keys equal to the specified property name
                            EXAMPLE, with props=['prop1', 'prop2]:
                                [{'prop1': 'value', 'prop2': 'value'}, ...]
        """
        assert len(props) > 0, 'Must specify at least one property to return!'
        assert len(props) == len(set(props)), 'Specified props must not contain duplicates!'

        q_return = f"RETURN c.`{props[0]}` as `{props[0]}`"
        if len(props) != 1:
            for prop in props[1:]:
                q_return += f", c.`{prop}` as `{prop}`"

        q = f"""
        MATCH (c:Class)
        {q_return}
        """

        return self.query(q)

    def get_rels_where(self, where_clause=None, return_prop="label") -> [{}]:
        """
        Returns a list of dictionaries representing relationships between all classes or a subset of classes and/or
        relationships if filtered with a cypher where clause.
        :param where_clause: Optional string cypher where clause, For example:
                             `WHERE from_class.short_label IS NOT NULL` - Only relationships from_classes with a short_label
        :param return_prop: Property to identify class in returned relationships. For example with
                            return prop = "label", relationships will be in the format:
                            {from: from_class.label, to: to_class.label, type: rel.relationship_type, optional: rel.relationship_optional}
        :return: A list of dicts representing relationships eg:
                            [{from: from_class.label, to: to_class.label, type: rel.relationship_type, optional: rel.relationship_optional}, ...]
        """

        q = f"""
        MATCH (from_class:Class)<-[:FROM]-(rel:Relationship)-[:TO]->(to_class:Class)
        {where_clause if where_clause else ""}
        RETURN {{from: from_class.`{return_prop}`, to: to_class.`{return_prop}`, type: rel.relationship_type, optional: rel.relationship_optional}} as rel   
        """
        res = self.query(q)
        return [x['rel'] for x in res]

    def get_rels_from_labels(self, labels: list) -> [{}]:
        """
        Returns all the relationships (according to the schema) from the nodes with specified labels
        including the relationships of parent and child classes
        """
        q = f"""
        MATCH 
            (c1:Class)<-[:SUBCLASS_OF*0..{str(self.SCD)}]-(c1low:Class)
        WHERE 
            c1.label in $labels AND
            NOT EXISTS ( (c1low)<-[:SUBCLASS_OF]-(:Class) )
        WITH c1low
        MATCH 
            path1 = (c1high:Class)<-[:SUBCLASS_OF*0..{str(self.SCD)}]-(c1low:Class)
        WHERE             
            NOT EXISTS ( (:Class)<-[:SUBCLASS_OF]-(c1high) )
        WITH nodes(path1) as col1
        UNWIND col1 as c1       
        MATCH (x)<-[f:FROM]-(rr:Relationship)-[t:TO]->(y)
        WHERE x = c1 or y = c1
        RETURN {{from: x.label, to: y.label, type: rr.relationship_type, optional: rr.relationship_optional}} as rel            
        ORDER BY rel['from'], rel['to'], rel['type']
        """
        params = {'labels': labels}
        res = self.query(q, params)
        return [x['rel'] for x in res]

    def get_labels_from_rels_list(self, rels_list: list) -> [str]:
        "Returns all the class labels from a list of relationships in the form {'from':...,'to':...,'type':...}"
        labels = []
        for rel in rels_list:
            for key in ['from', 'to']:
                if rel.get(key) not in labels:
                    labels.append(rel.get(key))
        return labels

    def get_rels_btw2(self, label1: str, label2: str, identifier='label'):
        """
        Returns all the relationships (according to the schema) between classes with identifier properties equal to
        {label1} and {label2} including the relationships of parent and child classes.
        """
        q = f"""
        MATCH 
            (c1:Class)<-[:SUBCLASS_OF*0..{str(self.SCD)}]-(c1low:Class),
            (c2:Class)<-[:SUBCLASS_OF*0..{str(self.SCD)}]-(c2low:Class)
        WHERE 
            c1.`{identifier}` = $label1 AND c2.`{identifier}` = $label2 AND
            NOT EXISTS ( (c1low)<-[:SUBCLASS_OF]-(:Class) ) AND
            NOT EXISTS ( (c2low)<-[:SUBCLASS_OF]-(:Class) ) 
        WITH c1low, c2low
        MATCH 
            path1 = (c1high:Class)<-[:SUBCLASS_OF*0..{str(self.SCD)}]-(c1low:Class),
            path2 = (c2high:Class)<-[:SUBCLASS_OF*0..{str(self.SCD)}]-(c2low:Class)
        WHERE             
            NOT EXISTS ( (:Class)<-[:SUBCLASS_OF]-(c1high) ) AND
            NOT EXISTS ( (:Class)<-[:SUBCLASS_OF]-(c2high) ) 
        WITH nodes(path1) as col1, nodes(path2) as col2
        UNWIND col1 as c1
        UNWIND col2 as c2
        WITH c1, c2
        MATCH (x)<-[f:FROM]-(rr:Relationship)-[t:TO]->(y)
        WHERE (x = c1 and y = c2) or (y = c1 and x = c2) 
        RETURN {{from: x.`{identifier}`, to: y.`{identifier}`, type: rr.relationship_type}} as rel            
        ORDER BY rel['from'], rel['to'], rel['type']
        """
        params = {'label1': label1, 'label2': label2}
        res = self.query(q, params)
        return [x['rel'] for x in res]

    def infer_rels(self, labels: list, oclass_marker: str = "**", impute_relationship_type: bool = True):
        """
        Infers most appropriate relationship type (if exists) between each pair of $labels
        for generating cypher query according to the schema
        """
        q = f"""
        MATCH (a:Class), (b:Class)
        WHERE a.label in $labels and b.label in $labels 
        AND (
                ( EXISTS ( (a)-[:SUBCLASS_OF*0..{str(
            self.SCD)}]->()<-[:FROM]-(:Relationship)-[:TO]->()<-[:SUBCLASS_OF*0..{str(self.SCD)}]-(b) ) )
                OR
                ( EXISTS ( (a)<-[:SUBCLASS_OF*0..{str(
            self.SCD)}]-()<-[:FROM]-(:Relationship)-[:TO]->()-[:SUBCLASS_OF*0..{str(self.SCD)}]->(b) ) )
            )
        OPTIONAL MATCH 
            p1 = (a)-[:SUBCLASS_OF*0..{str(
            self.SCD)}]->()<-[:FROM]-(r1:Relationship)-[:TO]->(crt_to1)<-[:SUBCLASS_OF*0..{str(self.SCD)}]-(b)
        OPTIONAL MATCH
            p2 = (a)<-[:SUBCLASS_OF*0..{str(
            self.SCD)}]-()<-[:FROM]-(r2:Relationship)-[:TO]->(crt_to2)-[:SUBCLASS_OF*0..{str(self.SCD)}]->(b)
        WITH a, b, 
            collect(distinct {{
                path: p1, 
                rel_type: 
                    CASE WHEN r1.relationship_type is NULL and $impute_relationship_type THEN 
                        crt_to1.label
                    ELSE
                        r1.relationship_type
                    END,   
                short_label: r1.short_label, 
                tag: 'p1'
            }})
            +
            collect({{
                path: p2, 
                rel_type: 
                    CASE WHEN r2.relationship_type is NULL THEN 
                        crt_to2.label
                    ELSE
                        r2.relationship_type
                    END,   
                short_label: r2.short_label,               
                tag: 'p2'
            }})
            as coll
        UNWIND coll as map
        WITH a, b, map, size(nodes(map['path'])) as sz 
        WHERE not map['path'] is NULL
        WITH *
        ORDER BY map['tag'], sz //Here we prioritizing paths with SUBCLASS > over paths with < SUBCLASS rel. and then prioritize shorter paths  
        WITH a, b, collect(map) as coll //ideally we should always have only 1 path for each pair, but allowing here for >1
        WITH a, b, coll[0] as map        
        //WITH apoc.map.mergeList(
        //    [{{from: a.label, to: b.label, type: map['rel_type']}}] +
        //        CASE WHEN short_label in keys(a) THEN [{{from_tag: a.short_label}}] ELSE [] END +
        //        CASE WHEN short_label in keys(b) THEN [{{to_tag: b.short_label}}] ELSE [] END
        //) as rel
        WITH {{from: a.label, to: b.label, type: map['rel_type'], short_label: map['short_label']}} as rel
        WITH *
        ORDER BY rel['from'], rel['to'], rel['type']        
        RETURN rel
        """
        olabels = [label for label in labels if label.endswith(oclass_marker)]
        params = {
            'labels': [(label[:-(len(oclass_marker))] if label in olabels else label) for label in labels],
            'impute_relationship_type': impute_relationship_type
        }
        res = self.query(q, params)
        rels = []
        for r in res:
            dct = r['rel']
            if dct.get('from') in olabels or dct.get('to') in olabels:
                dct['optional'] = True
            rels.append(dct)
        return rels

    def translate_to_shortlabel(self, labels: list, rels: list, labels_to_pack, where_map: dict = None,
                                where_rel_map: dict = None, use_rel_labels=True):
        if not where_map:
            where_map = {}
        if not where_rel_map:
            where_rel_map = {}

        q = """
        MATCH (c:Class)
        WHERE c.label in $labels
        RETURN apoc.map.fromPairs(collect([
            c.label,
            CASE WHEN 'short_label' in keys(c) THEN  
                c.short_label
            ELSE
                c.label
            END
        ])) as map
        """
        params = {'labels': list(set(labels + [rel.get('from') for rel in rels] + [rel.get('to') for rel in rels]))}
        assert labels_to_pack is None or isinstance(labels_to_pack, dict)
        if labels_to_pack is not None:
            labels_lst = []
            for key, value in labels_to_pack.items():
                assert isinstance(value,
                                  (str, list)), f'Value in labels_to_pack is not string or list. It was: {type(value)}'
                labels_lst.append(key)
                if isinstance(value, str):
                    labels_lst.append(value)
                elif isinstance(value, list):
                    labels_lst.extend(value)
            params['labels'].extend(labels_lst)
        # print(f'PARAMS: {params}')

        res = self.query(q, params)
        map = res[0]['map']
        if use_rel_labels:
            for rel in rels:
                if rel.get('short_label'):
                    map[rel['to']] = rel.get('short_label')
        if not labels:
            labels = []
        labels = [{'label': label, 'short_label': map[label]} for label in labels]
        if not rels:
            rels = []
        rels = [{**rel, **{'from': map[rel['from']], 'to': map[rel['to']]}} for rel in rels]

        if labels_to_pack is not None:
            mapped_labels_to_pack = {}
            # print(f'MAP: {map}')
            for key, value in labels_to_pack.items():
                assert isinstance(value,
                                  (str, list)), f'Value in labels_to_pack is not string or list. It was: {type(value)}'
                if isinstance(value, str):
                    mapped_labels_to_pack[map[key]] = map[value]
                elif isinstance(value, list):
                    # print(f'KEY: {key}')
                    # print(f'VALUE: {value}')
                    assert key in map.keys(), f'{key} was not found as a key in {map}'
                    assert (True if i in map.keys() else False for i in
                            value), f'a value from {value} was not found as a key in {map}'
                    mapped_labels_to_pack[map[key]] = [map[i] for i in value]
                    # convert the fromclass/coreclass label to short label only
                    # leave the label as it needs to be long format for the generate_call function
            labels_to_pack = mapped_labels_to_pack

        if not where_map:
            where_map = {}
        where_map = {map[key]: item for key, item in where_map.items()}
        where_rel_map = {map[key]: item for key, item in where_rel_map.items()}
        return labels, rels, labels_to_pack, where_map, where_rel_map

    @staticmethod
    def arrows_dict_uri_dict_enrich(dct: dict, uri_map: dict):

        def _gen_new_prop_name(neighbour: dict):
            assert len({"label", "relationship", "property"}.intersection(neighbour.keys())) == 3
            return f"{neighbour['relationship']}.{neighbour['label']}.{neighbour['property']}"

        def _get_neighbour_id(nd: dict, neighbour: dict):
            assert len({"label", "relationship", "property"}.intersection(neighbour.keys())) == 3
            for rel in dct['relationships']:
                if nd["id"] in [rel["toId"]] and rel["type"] == neighbour["relationship"]:
                    return rel["fromId"]
                elif nd["id"] in [rel["fromId"]] and rel["type"] == neighbour["relationship"]:
                    return rel["toId"]
            return None

        def _get_neighbour_value(nd: dict, neighbour: dict):
            neighbour_id = _get_neighbour_id(nd, neighbour)
            for nd in dct['nodes']:
                if nd['id'] == neighbour_id:
                    return nd['properties'].get(neighbour["property"])

        def _enrich_node(nd: dict, neighbour: str):
            value = _get_neighbour_value(nd, neighbour)
            return {**nd, **{"properties": {**nd["properties"], **{_gen_new_prop_name(neighbour): value}}}}

        merge_on = {}
        for key, item in uri_map.items():
            merge_on[key] = item.get("properties").copy()
            if not merge_on[key]:
                merge_on[key] = []
            if item.get('neighbours'):
                for n in item['neighbours']:
                    new_nodes = []
                    for nd in dct['nodes']:
                        if key in nd['labels']:
                            new_nodes.append(_enrich_node(nd, n))
                        else:
                            new_nodes.append(nd)
                    dct = {**dct, **{"nodes": new_nodes}}
                    merge_on[key].append(_gen_new_prop_name(n))
        return dct, merge_on

    def create_ct(self, controlled_terminology: dict, identifier='label', order_terms=True, merge_on=None):
        """
        Creates :Term nodes and links them to a specified class with a [:HAS_CONTROLLED_TERM] relationship.
        If order terms is True, an ascending Order property will be assigned to terms in the order they are created
        accounting for the order of existing terms (if any) and [:NEXT] relationships between terms following this order.
        For example:
            With identifier = 'label' and controlled_terminology =
            {
                'class1': [{'term_label': 'term1'}, {'term_label': 'term2'}],
                'class2': [{'term_label': 'term3'}]
            }
            3 new term nodes with 'term_label' properties equal to 'term1', 'term2' and 'term3' would be
            created. The class with 'label' = 'class1' would then be linked by [:HAS_CONTROLLED_TERM]
            relationships to 'term1' and 'term2' and similarly 'class2' would be linked to 'term3'

        :param controlled_terminology: A dictionary of classes and terms eg: {'class1': [{prop1: value, prop2: value}, ...]}
        :param identifier: String, property used when identifying classes to assign terms.
        :param order_terms: Bool, if true order properties and next relationships will be created between terms.
        :param merge_on: Optional List[string] term properties to merge on to prevent duplication
        :return: neo4j result object.
        """
        ident_props = None
        missing = self.get_missing_classes(list(controlled_terminology.keys()), identifier)
        assert not missing, f'Cannot create controlled terminology for nonexistent classes: {missing}'

        if merge_on:
            ident_props = f'{{`{merge_on[0]}`: term_props["{merge_on[0]}"]'
            for prop in merge_on[1:]:
                ident_props += f', `{prop}`: term_props["{prop}"]'
            ident_props += '}'

        # Create terms
        q1 = f"""
        UNWIND KEYS($terminology) as class_label
        MATCH (class:Class {{{identifier}: class_label}})
        WITH class, class_label
        UNWIND $terminology[class_label] as term_props
        CALL apoc.merge.node(['Term'], {f"{ident_props}, term_props, term_props" if ident_props else 'term_props, {}, {}'}) YIELD node
        CALL apoc.create.addLabels([node], [class.label]) YIELD node as term
        MERGE (class)-[:HAS_CONTROLLED_TERM]->(term)
        """
        res1 = self.query(q1, {'terminology': controlled_terminology}, return_type='neo4j.Result')

        if order_terms:
            # Create order property on new terms
            q2 = f"""
            UNWIND $labels as class_label
            MATCH (c:Class{{{identifier}: class_label}})
            OPTIONAL MATCH (c)-[:HAS_CONTROLLED_TERM]->(term:Term)
            WITH c, MAX(term.Order) as term_order
            WITH c, CASE WHEN term_order IS NULL THEN 1 ELSE term_order + 1 END as term_order
    
            MATCH (c)-[:HAS_CONTROLLED_TERM]->(t1:Term)
            WHERE t1.Order is NULL
            WITH c, term_order, t1 order by t1.`Codelist Code`, t1.`Term Code` 
            WITH c, collect(t1) as terms_to_order, term_order
            WITH *, apoc.coll.zip(terms_to_order, range(term_order, term_order + size(terms_to_order))) as pairs
            UNWIND pairs as pair
            WITH pair[0] as term_node, pair[1] as new_order, c, terms_to_order
            SET term_node.Order = new_order
            return c, terms_to_order
            """
            self.query(q2, {'labels': list(controlled_terminology.keys())})

            # Create next rel between terms
            q3 = f"""
            UNWIND $labels as class_label
            MATCH (c:Class{{{identifier}: class_label}})-[:HAS_CONTROLLED_TERM]->(t:Term)
            WITH c,t ORDER BY c.label, t.Order ASC
            WITH c, COLLECT(t) AS terms
            FOREACH (n IN RANGE(0, SIZE(terms)-2) |
                FOREACH (prev IN [terms[n]] |
                    FOREACH (next IN [terms[n+1]] |
                        MERGE (prev)-[:NEXT]->(next))))
            """
            self.query(q3, {'labels': list(controlled_terminology.keys())})

        return res1

    def get_class_ct(self, class_: str, ct_prop_name='rdfs:label'):
        q = """
        MATCH (c:Class) 
        WHERE c.label = $class_
        OPTIONAL MATCH (c)-[:HAS_CONTROLLED_TERM]->(t:Term)
        RETURN collect(DISTINCT t[$ct_prop_name]) as coll
        """
        params = {'class_': class_, 'ct_prop_name': ct_prop_name}
        res = self.query(q, params)
        if res:
            return res[0]['coll']
        else:
            return []

    def get_class_ct_map(self, classes: list, ct_props: list = None, identifier='label'):
        """
        Return controlled terminology for a given list of classes.
        :param classes: list of class labels of class.identifier property values if using custom identifier
                        for example: with identifier = 'short_label', classes = ['shortlabel1', 'shortlabel2']
        :param ct_props: list of controlled terminology props to collect eg:
                         ['label', 'Codelist Code', 'Order']
                         defaults to ['rdfs:label']
        :param identifier: string used when identifying classes
        :return: Dictionary of classes and found controlled terminology for example:
                 with classes = ['class1', 'class2'] and ct_props = ['label'] a typical result might be:
                {'class1': [{'label': 'term1'}, {'label': 'term2'}], 'class2': [{{'label': 'term3'}}]}
                Where 'class1' has two terms labeled 'term1' and 'term2' and 'class2' has one term labeled 'term3'
        """

        if ct_props is None:
            ct_props = ['rdfs:label']
        elif type(ct_props) == str:
            ct_props = [ct_props]

        prop_collection = f'["{ct_props[0]}", term.`{ct_props[0]}`]'
        if len(ct_props) != 1:
            for prop in ct_props[1:]:
                prop_collection += f', ["{prop}", term.`{prop}`]'

        q = f"""
        UNWIND $classes as class
        MATCH (c:Class)-[:HAS_CONTROLLED_TERM]->(term:Term)
        WHERE c.`{identifier}` = class
        WITH c, collect(apoc.map.fromPairs([{prop_collection}])) as terms
        RETURN apoc.map.setKey({{}}, c.`{identifier}`, terms) as ct
        """

        params = {'classes': classes}
        res = self.query(q, params)

        data = {}
        if res:
            for term_dict in res:
                # Flattened result {class.identifer:[[term.ct_props[0], ...], [term.ct_props[0], ...]]}
                data.update(term_dict.get('ct'))
        return data

    def delete_ct(self, controlled_terminology: dict, ct_props: list, identifier='label'):
        """
        Deletes part of class specific controlled terminology.
        :param controlled_terminology: Dictionary of key: class identity - value: list of class specific term property
                                       values to delete. For example:
                                       with ct_props = ['Codelist Code'] controlled terminology might be:
                                      {'class1':[['code1'], ['code2']], class1':[['code2']]}
                                      Note these property values must align with the keys defined in term_props!
        :param ct_props: List of property names that define property values in a controlled terminology term,
                         For example:
                            With controlled terminology = {'class1':[['code1'], ['code2']], class1':[['code2']]}
                            ct_props would be ['Codelist Code'] indicating that we intending to delete terms
                            for 'class1' and 'class2' using 'Codelist Code' property value.
        :param identifier: string, property used to identify class
        :return: neo4j result object.
        """
        # TODO: Resolving [:NEXT] rel when deleting CT

        where_clause = f't.`{ct_props[0]}` = term_props[0]'
        for count, prop in enumerate(ct_props[1:], start=1):
            where_clause += f' AND t.`{prop}` = term_props[{count}]'

        q = f"""
        UNWIND KEYS($terminology) as class_label
        MATCH (c:Class)
        WHERE c.`{identifier}` = class_label
        UNWIND $terminology[class_label] as term_props
        MATCH (c)-[:HAS_CONTROLLED_TERM]-(t:Term)
        WHERE {where_clause}
        DETACH DELETE t
        """

        return self.query(q, {'terminology': controlled_terminology}, return_type='neo4j.Result')

    def get_all_ct(self, term_props: list, class_prop='label', derived_only=False):
        """
        Returns a list of dictionaries containing specified props for all controlled terminology.
        Example:
            With term_props = ['rdfs:label'] and class_prop = ['label'] the return format would be:
            [{'label': class1, 'rdfs:label': 'term1'}, {'label': class1, 'rdfs:label': 'term2'},
            {'label': class2, 'rdfs:label': 'term1'} ...]
            Repeated for all CT.
        :param term_props: Term properties to return eg: ['rdfs:label', 'Codelist Code']
        :param class_prop: Class property used to identify controlled terminology class
                            :param derived_only:
        :param derived_only: Bool, optional filter for classes where class.derived = 'true'.
        :return: List of dictionaries, see example above.
        """
        # TODO: consider converting to format similar to get_class_ct_map? ie {'class.label': [{'rdfs:label': 'term1'}]}
        assert len(term_props) >= 1, 'Must include at least 1 term_prop'
        assert class_prop not in term_props, 'Class prop cannot be in term props'

        term_return = f'term.`{term_props[0]}` as `{term_props[0]}`'
        for prop in term_props[1:]:
            term_return += f', term.`{prop}` as `{prop}`'

        q = f"""
        MATCH (c:Class)-[:HAS_CONTROLLED_TERM]->(term:Term)
        {'WHERE c.derived = "true"' if derived_only else ''}
        RETURN c.`{class_prop}` as `{class_prop}`, {term_return}
        """
        return self.query(q)

    def create_same_as_ct(self, same_as_terms: List[dict], term_identifiers: List[str], identifier='label'):
        """
        Creates a [:SAME_AS] relationship between two terms. For example:
            With term_identifiers = ['Codelist Code', 'Term Code']
            and same_as_terms = [
                {'from_class': 'class1', 'to_class': 'class2', # Class identifiers
                 'from_codelist_code': 'code_1', 'from_term_code': 'code_2', # Term 1 identifier
                 'to_codelist_code': 'code_3', 'to_term_code': 'code_4' # Term 2 identifier
                }
            ]
            A same as relationship would be created from a term of 'class1' with properties:
            `Codelist Code` = code_1 and `Term Code` = 'code_2' to the corresponding term for 'class2'.

            Note term identifiers in a same_as_terms dictionaries must be lowercase, use underscores instead of spaces
            and be prefixed with from_ or to_ a property listed in term_identifiers.

        :param same_as_terms: List of dictionaries defining same as terms between classes.
        :param term_identifiers: List of strings with term properties that guarantee uniqueness
        :param identifier: string class property used when matching classes
        :return: neo4j result object
        """

        where_clause = f"WHERE c1.`{identifier}` = new_term['from_class'] AND c2.`{identifier}` = new_term['to_class']"
        for term_prop in term_identifiers:
            clean_prop = term_prop.lower().replace(' ', '_')
            where_clause += f" AND t1.`{term_prop}` = new_term['from_{clean_prop}']"
            where_clause += f" AND t2.`{term_prop}` = new_term['to_{clean_prop}']"

        q = f"""
        UNWIND $same_as_terms as new_term
        MATCH (c1:Class)-[:HAS_CONTROLLED_TERM]-(t1:Term)
        MATCH (c2:Class)-[:HAS_CONTROLLED_TERM]-(t2:Term)
        {where_clause}
        MERGE (t1)-[:SAME_AS]->(t2)
        """
        return self.query(q, {'same_as_terms': same_as_terms}, return_type='neo4j.Result')

    def remove_same_as_ct(self, same_as_terms: List[dict], term_identifiers: List[str], identifier='label'):
        """
        Removes a [:SAME_AS] relationship between two terms, see "create_same_as_ct()" for format information.

        :param same_as_terms: List of dictionaries defining same as terms between classes.
        :param term_identifiers: List of strings with term properties that guarantee uniqueness
        :param identifier: string class property used when matching classes
        :return: neo4j result object
        """

        where_clause = f"WHERE c1.`{identifier}` = new_term['from_class'] AND c2.`{identifier}` = new_term['to_class']"
        for term_prop in term_identifiers:
            clean_prop = term_prop.lower().replace(' ', '_')
            where_clause += f"AND t1.`{term_prop}` = new_term['from_{clean_prop}']"
            where_clause += f"AND t2.`{term_prop}` = new_term['to_{clean_prop}']"

        q = f"""
        UNWIND $same_as_terms as new_term
        MATCH (c1:Class)-[:HAS_CONTROLLED_TERM]->(t1:Term)-[rel:SAME_AS]->(t2:Term)<-[:HAS_CONTROLLED_TERM]-(c2:Class)
        {where_clause}
        DETACH DELETE rel
        """

        return self.query(q, {'same_as_terms': same_as_terms}, return_type='neo4j.Result')

    def propagate_rels_to_parent_class(self):
        if self.verbose:
            logger.info("Copying Relationships to 'parent' Classes where (child)-[:SUBCLASS_OF]->(parent)")
        self.query("""
        MATCH (c:Class)<-[r1:TO|FROM]-(r:Relationship)-[r2:TO|FROM]-(target:Class), (c)-[:SUBCLASS_OF*1..50]->(source:Class)
        WHERE type(r1) <> type(r2)
        WITH *,
        "
            WITH $source as source, $target as target
            MERGE (source)<-[:`"+type(r1)+"`]-(:Relationship{relationship_type:$type})-[:`"+type(r2)+"`]->(target)
            RETURN count(*)
        " as q, 
        {type: r.relationship_type, source: source , target: target} as params
        CALL apoc.cypher.doIt(q, params) YIELD value
        RETURN value, q, params         
        """)

    def remove_unmapped_classes(self):
        if self.verbose:
            logger.info("Removing Unmapped Classes")
        q = """
        MATCH (c:Class)
        WHERE NOT EXISTS (
            ()-[:MAPS_TO_CLASS]->()-[:SUBCLASS_OF*0..50]->(c)
        )
        OPTIONAL MATCH (c)<-[:TO|FROM]-(r:Relationship)
        REMOVE r:Relationship
        REMOVE c:Class
        """
        self.query(q)

    def remove_auxilary_term_labels(self):
        if self.verbose:
            logger.info("Removing Auxilary Term Labels")
        """
        To be used after reshaping - additional labels from Terms that have not been extracted from data are removed
        :return:
        """
        q = """
        MATCH (x:Term) 
        WHERE not exists ( (x)-[:FROM_DATA]->() )
        WITH x, labels(x) as coll
        UNWIND coll as lbl
        WITH *
        WHERE NOT lbl in ["Term", "Class"]
        CALL apoc.cypher.doIt(
        'WITH x REMOVE x:`'+lbl+'`',
        {x:x}
        ) yield value
        RETURN value
        """
        self.query(q)

    def export_model_ttl(self, folder: str, filename: str, include_mappings=False):
        uri_map1 = {
            "Data Extraction Standard": {"properties": "_tag_"},
            "Source Data Folder": {"properties": "_folder_"},
            "Source Data Table": {"properties": "_domain_"},
            "Source Data Column": {"properties": ["_domain_", "_columnname_"]}
        }
        uri_map2 = {key: item
                    for key, item in ModelManager.URI_MAP.items()
                    if key in ["Class", "Property", "Relationship"]}

        if include_mappings:
            self.rdf_generate_uri({**uri_map1, **uri_map2})
            rdf = self.rdf_get_subgraph(
                """
                MATCH (n:Class)
                call apoc.path.expand(n, 
                    'SUBCLASS_OF|HAS_CONTROLLED_TERM>|<MAPS_TO_CLASS|<HAS_COLUMN|<HAS_TABLE|<TO|<FROM', 
                    '+Class|+Term|+Relationship|+Data Extraction Standard|+Source Data Folder|+Source Data Table|+Source Data Column', 
                    0, 1 ) yield path
                RETURN path
                """
            )
        else:
            self.rdf_generate_uri(uri_map2)
            rdf = self.rdf_get_subgraph(
                """
                MATCH (n:Class)
                call apoc.path.expand(n, 'SUBCLASS_OF|HAS_CONTROLLED_TERM|<TO|<FROM', '+Class|+Relationship|+Term', 0, 1 ) yield path
                RETURN path
                """
            )
        rdf_file_path = os.path.join(folder, filename)
        with open(rdf_file_path, "w") as file:
            file.write(rdf)

    def create_custom_mappings_from_dict(self, groupings = None) -> None:
        """
        Function to support MethodApplier(mode="schema_PROPERTY")

        Create a set of nodes labeled "Class", all taken from the groupings dictionary; their names are
            stored in an attribute named "label".  EXAMPLES: "Study", "Site", "Race", "Adverse Event".

        Also, create a set of nodes labeled "Property"; likewise, their names are
            stored in an attribute named "label".  EXAMPLES: "RACE", "RACEN", "SITEID".

        In addition, create relationships named "HAS_PROPERTY", from each of the "Class" nodes to the appropriate
            "Property" nodes.   EXAMPLE:  The "Race" node (labeled "Class") links to the "RACE", "RACEN" nodes
            (labeled "Property")

        Finally,  create relationships named "MAPS_TO_PROPERTY", from `Source Data Column` nodes to `Property` nodes.
            These relationships are created based on matches between the "_columnname_" attribute on `Source Data Column` nodes
            and entries in the lists contained in groupings (e.g., "RACEN"); in some cases, further restrictions are applied,
            requiring the `Source Data Column` node to be linked to a particular `Source Data Table` node.

        :param groupings:   A dictionary.  The keys are either "*" (meaning no `Source Data Table` restriction)
                                           or the name of a specific `Source Data Table`.
                                           The values are dictionaries such as {"Race": ["RACE", "RACEN"]}
        :return:            None
        """
        if not groupings:
            groupings = {}

        # Loop over all the keys/values of the groupings dictionary
        for table, groupping in groupings.items():
            # EXAMPLES of table:  "ADSL" or "*" (the star indicates "all tables")
            # groupping is a dictionary with entries such as  "Race": ["RACE", "RACEN"]

            # Define the Cypher query, which depends on the value of the "table" variable in the outer loop;
            #       for examples of how the query shapes up, see below (inside inner loop)
            q = f"""                    
                MERGE (class:Class {{label:$class}})
                WITH * UNWIND $properties as property
                MERGE (class)-[:HAS_PROPERTY]->(p:Property {{label:property}})
                WITH *
                MATCH (sdc:`Source Data Column` {{_columnname_:property}})
                {
                "" if table == "*" else "<-[:HAS_COLUMN]-(sdt:`Source Data Table` {_domain_:$table})"
                }                     
                MERGE (sdc)-[:MAPS_TO_PROPERTY]->(p)
                """

            for class_, properties in groupping.items():
                # EXAMPLE of class_ : "Race"
                # EXAMPLE of properties: ["RACE", "RACEN"]

                params = {"table":table, "class": class_, "properties": properties}
                self.query(q, params)

                # EXAMPLE1 of query (involving a specific table, such as "ADSL")
                """
                    MERGE (class:Class {label:$class})
                    WITH * UNWIND $properties as property
                    MERGE (class)-[:HAS_PROPERTY]->(p:Property {label:property})
                    WITH *
                    MATCH (sdc:`Source Data Column` {_columnname_:property})
                    <-[:HAS_COLUMN]-(sdt:`Source Data Table` {_domain_:$table})                     
                    MERGE (sdc)-[:MAPS_TO_PROPERTY]->(p)
                """

                # EXAMPLE2 of query (involving any table, indicated by the "*" value of the table variable)
                """
                    MERGE (class:Class {label:$class})
                    WITH * UNWIND $properties as property
                    MERGE (class)-[:HAS_PROPERTY]->(p:Property {label:property})
                    WITH *
                    MATCH (sdc:`Source Data Column` {_columnname_:property})                     
                    MERGE (sdc)-[:MAPS_TO_PROPERTY]->(p)
                """
                # EXAMPLE of params: {'table': 'ADSL', 'class': 'Race', 'properties': ['RACE', 'RACEN']}

    def create_custom_rels_from_list(self, rels: [[str, str]], create_if_absent = False) -> None:
        """
        Adds "CLASS_RELATES_TO" relationships to pairs of nodes with "Class" label,
        based on the list passed as the argument "rels".
        Matches occur based on node attributes named "label".
        Optionally, create the Class nodes if needed.
        To create just a single relationship, see create_class_relationship()

        NOTE: this is a more general version of the method create_related_classes_from_list()

        :param rels:    A list of 2-element lists, indicating a relationship among nodes of type `Class`
                        EXAMPLE:   [
                                        ["Study", "Site"],
                                        ["Study", "Subject"],
                                        ["Subject", "Race"]
                                   ]
        :param create_if_absent: If True, the Class nodes specified in the argument "rel" get created as needed;
                                        otherwise, no relationships get created whenever their start or end class is missing
        :return:                 None
        """
        if rels is None or rels == []:
            return      # There's nothing to do

        if create_if_absent:
            q = f"""            
                UNWIND $rels as rel
                WITH rel[0] as left, rel[1] as right      
                WHERE apoc.meta.type(left) = apoc.meta.type(right) = 'STRING'  
                MERGE (ln:Class {{label:left}})
                MERGE (rn:Class {{label:right}})   
                MERGE (ln)-[:CLASS_RELATES_TO]->(rn)   
                """
        else:   # The Class nodes MUST be present, or no relationship gets created
            q = f"""            
                UNWIND $rels as rel
                WITH rel[0] as left, rel[1] as right    
                WHERE apoc.meta.type(left) = apoc.meta.type(right) = 'STRING'      
                MATCH (ln:Class {{label:left}}), (rn:Class {{label:right}})    
                MERGE (ln)-[:CLASS_RELATES_TO]->(rn)   
                """

        params = {"rels": rels}
        self.query(q, params)

    def create_model_from_data(
            self,
            data_label: str = "Source Data Row",
            data_table_label: str = "Source Data Table",
            domain_property: str = "_domain_",
            no_domain_label: str = "Thing",
            data_column_label: str = "Source Data Column",
            columnname_property: str = "_columnname_",
            exclude_properties: list = None,
    ):
        """
        Creates Class and Relationship nodes to represent a trivial schema to reshape data ingested e.g. with
        FileDataLoader - Tables loaded into nodes one row to one node, column names used as property names.

        data_labels: labels of the nodes where loaded data is stored (mm with use OR btw labels to fetch data nodes)
        domain_property: property where the name of the table/domain can be found
        """
        if exclude_properties is None:
            exclude_properties = ["_filename_", "_folder_"]
        q = f"""
        MATCH (data:`{data_label}`)<-[:HAS_DATA]-(dt:`{data_table_label}`)        
        WITH distinct dt, dt._domain_ as domain, keys(data) as ks
        WITH *, CASE WHEN domain IS NULL THEN $no_domain_label ELSE domain END as domain
        WITH dt, domain, [k in ks WHERE NOT k IN $exclude_properties] as ks
        WITH dt, domain, apoc.coll.flatten(apoc.coll.toSet(ks)) as ks
        MERGE (c:Class{{label: domain, short_label: domain, create: True}})
        MERGE (dt)-[:MAPS_TO_CLASS]->(c)
        WITH *
        UNWIND ks as k
        MATCH (dt)-[:HAS_COLUMN]->(col:`{data_column_label}`)
        WHERE col.`{columnname_property}` = k
        MERGE (c2:Class{{label: k, short_label: k}})        
        MERGE (c)<-[:FROM]-(r:Relationship{{relationship_type: k}})-[:TO]->(c2)
        MERGE (col)-[:MAPS_TO_CLASS]->(c2)        
        """
        params = {
            "exclude_properties": exclude_properties + [domain_property],
            "no_domain_label": no_domain_label
        }
        self.query(q, params)


from derivation_method.action import GetData
from derivation_method.derivation_method import OnlineDerivationMethod, derivation_method_factory, DictDerivationMethod, RdfDerivationMethod, SnippetDerivationMethod, DerivationMethod



import json
import logging
import time
import os
from functools import reduce
from operator import add
from pathlib import Path
from typing import List

from neointerface import NeoInterface

from data_providers import DataProvider
from derivation_method.action import GetData, CallAPI, RunScript, Link, AssignLabel, BuildUri
from derivation_method.method import Method
from derivation_method.super_method import SuperMethod, ApplyStatSuperMethod, DecodeSuperMethod, SubjectLevelLinkSuperMethod
from derivation_method.utils import simplify_arrows_json
from logger.logger import logger
from model_managers import ModelManager
from derivation_method.utils import visualise_json, topological_sort

RDFSLABEL = ModelManager.RDFSLABEL


def derivation_method_factory(interface,
                              data="online",
                              study=None,
                              name=None,
                              overwrite_db=False,
                              schema_available=True,
                              check_schema=False):
    if isinstance(data, dict):
        return DictDerivationMethod(content=data,
                                    name=name,
                                    interface=interface,
                                    study=study,
                                    overwrite_db=overwrite_db,
                                    schema_available=schema_available,
                                    check_schema=check_schema)
    elif isinstance(data, Path) or isinstance(data, str) and data.endswith(".json") or data.endswith(".ttl"):
        path = Path(data)
        if path.suffix == ".json":
            with path.open() as file:
                return DictDerivationMethod(content=json.load(file),
                                            interface=interface,
                                            study=study,
                                            name=path.name.replace(path.suffix, ""),
                                            overwrite_db=overwrite_db,
                                            schema_available=schema_available,
                                            check_schema=check_schema)
        elif path.suffix == ".ttl":
            with path.open() as file:
                return RdfDerivationMethod(content=file.read(),
                                           interface=interface,
                                           study=study,
                                           name=path.name.replace(path.suffix, ""),
                                           overwrite_db=overwrite_db,
                                           schema_available=schema_available)
        else:
            if 'archive' not in path.name:
                raise TypeError(f"Unknown File Type {path.suffix} for {path.name}")

    else:
        return OnlineDerivationMethod(name=data, interface=interface, study=study)


class DerivationMethod(Method):  # common variables
    """
    Metadata for Derivations
    """

    def __init__(self, data, interface: NeoInterface = None, name=None, study=None, schema_available=None):
        super().__init__()
        self.interface = interface if interface is not None else DataProvider(rdf=True)
        self.schema_available = True if schema_available is None else schema_available
        self.batch = False
        self.data = data
        self.name = name
        self._db_id = None
        self._actions = None
        if study is None:
            res = self.interface.query(f"""
            MATCH (study:Study) 
            RETURN study.`{RDFSLABEL}` as STUDYID 
            ORDER BY STUDYID""")
            assert len(
                res) == 1, "Provide 'study' id at init of DerivationMethod, as there are 0 or >1 studies in the database"
            if res:
                study = res[0]['STUDYID']
        self.study = study

    @property
    def exists_in_db(self) -> bool:
        q = """
            MATCH (m:Method{id:$method_id, parent_id:$study_id})
            RETURN id(m)
            """
        params = {'method_id': self.name, 'study_id': self.study}
        return len(self.interface.query(q, params)) > 0

    @property
    def db_id(self):
        if self._db_id is None:
            q = """
                MATCH (m:Method{id:$method_id})
                WHERE NOT EXISTS (m.parent_id) or (m.parent_id) = $study_id
                RETURN id(m)
                """
            params = {'method_id': self.name, 'study_id': self.study}
            self._db_id = self.interface.query(q, params)[0]["id(m)"]
        return self._db_id

    @property
    def actions(self):
        raise NotImplementedError

    @property
    def applied(self):
        if self.name == "online":
            return False
        q = """
            MATCH (m:Method{id:$method_id, parent_id:$study_id})-[r:METHOD_ACTION]-(action:Method),
            (action)-[:APPLIED]->(chg:Changes)
            RETURN distinct chg.applied_datetime
            """
        params = {'method_id': self.name, 'study_id': self.study}
        res = self.interface.query(q, params)
        return len(res) > 0

    def enrich_existing_relationship_nodes(self):
        # Enriching existing Relationship nodes so that load_method_json_inline merges correctly
        q = """
        MATCH (r:Relationship)-[:FROM]->(from), (r)-[:TO]->(to)
        SET r.`FROM.Class.label` = from.label
        SET r.`TO.Class.label` = to.label
        """
        self.interface.query(q)

    def load(self):
        logger.info(f"Loading {self.name}")

    def load_decision(self, overwrite_db) -> bool:
        if self.exists_in_db:
            if overwrite_db:
                return True
            else:
                _ = self.db_id
                logger.info(
                    f"Derivation {self.name} already exists in neo4j (node id: {self.db_id}), method will not be reloaded. "
                    f"To overwrite the version already in the database, set overwrite_db=True")
                return False
        else:
            return True

    def post_load_enrichment(self):
        new_method_node_ids = self.get_new_method_node_ids()  # result like [{'id(m)': 211325}]
        for qres in new_method_node_ids:
            self.set_nodes_parent_id(method_node_id=qres['id(m)'])

        # NOTE: here we regenerate uri to include parent_id(study id) in the uri this becomes an instance of
        # the selected method specifically for the study(or pooled analysis) - i.e. database
        if self.schema_available:  # limit URI generation to the newly loaded method nodes
            self.interface.rdf_generate_uri({key: item
                                             for key, item in ModelManager.URI_MAP.items()
                                             if key in ["Method"]
                                             })
            self.explicit_inputs_of_method()
            self.explicit_outputs_of_method()

        else:  # need to generate URIs on all nodes loaded with the method
            self.interface.rdf_generate_uri(ModelManager.URI_MAP)

        if not self.batch:
            _ = self.db_id  # get db_id

    def build_retrieve_actions_params(self):
        return {"method_id": self.name, "parent_id": self.study}

    def handle_supermethod(self, action_info, previous_actions=None):
        previous_actions = previous_actions if previous_actions is not None else []
        if action_info["type"] == "apply_stat":
            return ApplyStatSuperMethod(method=self, interface=self.interface, action_info=action_info, previous_actions=previous_actions)
        if action_info["type"] == "decode":
            return DecodeSuperMethod(method=self, interface=self.interface, action_info=action_info, previous_actions=previous_actions)
        if action_info["type"] == "subject_level_link":
            return SubjectLevelLinkSuperMethod(method=self, interface=self.interface, action_info=action_info, previous_actions=previous_actions)
        return SuperMethod(method=self, interface=self.interface, action_info=action_info)

    def check_method_actions_order(self):
        for i, action in enumerate(self.actions):
            if isinstance(action, GetData):
                if i == len(self.actions) - 1 or isinstance(self.actions[i + 1], GetData):
                    raise TypeError(f"get_data cannot be last or followed by get_data Method Action (index {i})")
            
    def get_new_method_node_ids(self):
        q = """
        MATCH (m:Method{id:$method_id})
        WHERE NOT EXISTS (m.parent_id)
        RETURN id(m)
        """
        params = {"method_id": self.name}
        return self.interface.query(q, params)

    def set_nodes_parent_id(self, method_node_id):
        q = """
        MATCH (m:Method)
        WHERE id(m) = $method_node_id
        SET m.parent_id = $parent_id
        WITH *
        MATCH (m)-[:METHOD_ACTION*1..2]->(child:Method)
        RETURN id(child), m.id
        """
        params = {"method_node_id": method_node_id, "parent_id": self.study}
        results = self.interface.query(q, params)
        for qres in results:
            self.interface.query(q, {"method_node_id": qres["id(child)"],
                                     "parent_id": "_".join([self.study, qres["m.id"]])})

    def apply_generator(self, set_branch_dict: dict = None, limit: int = 0, apply_changes: bool = True):
        if set_branch_dict is None:
            set_branch_dict = {}
        if not self.applied:
            self.check_method_actions_order()
            logger.info(f"Applying {self.name} for Study {self.study}")
            logger.info(f"\tTotal # of Actions: {len(self.actions)}")
            df = None
            for i, action in enumerate(self.actions):
                logger.info(f"Applying {action}: #{i + 1} of {len(self.actions)} of ({self.name})")
                start_time = time.time()

                if isinstance(action, GetData):
                    df = action.apply(df, limit=limit)

                elif isinstance(action, CallAPI) and (action.meta.get('github_repo') in set_branch_dict):
                    # when a CallAPI action has a github_repo that is found in set_branch_dict, we use that branch for
                    # the API call
                    repo = action.meta.get('github_repo')
                    df = action.apply(df, github_branch=set_branch_dict[repo])

                elif isinstance(action, (Link, AssignLabel)):
                    if apply_changes:
                        df = action.apply(df)
                    else:
                        logger.info(f'\tSkipping action as applying in limited mode')
                        continue
                else:
                    df = action.apply(df)
                logger.debug(f"\t\tCompleted in: {(time.time() - start_time):.2f}' seconds")
                if df is None:
                    logger.warning(f'Action {action} returned a None df!')
                yield df

    def apply(self, overwrite_derived_data=None, set_branch_dict: dict = None, limit: int = 0, apply_changes=True):
        """
        :param set_branch_dict: Dictionary of form {'repo': 'branch', ...} to allow the CallAPI action to use branches
        that are not main
        :param overwrite_derived_data: Bool
        :param limit: int, number of rows to return where 0 is unlimited
        :param set_branch_dict: dict, a dictionary of {'github_branch': branch, 'github_token': token, 'github_baseurl': baseurl} to be passed to call_api actions if required
        :param apply_changes: Bool
        :return:
        """
        if set_branch_dict is None:
            set_branch_dict = {}
        if self.applied:
            if overwrite_derived_data:
                OnlineDerivationMethod(name=self.name, interface=self.interface).rollback()
            else:
                logger.warning(
                    f"Method {self.name} has been derived, to reapply, either rollback or set overwrite_derivation_data=True"
                    f"Continuing...")
                return logger.info(f"Continuing...")
        return list(self.apply_generator(set_branch_dict, limit, apply_changes))[-1]

    def apply_limited(self, overwrite_derived_data=None, set_branch_dict: dict = None, limit: int = 10):
        logger.info(f'Running apply in limited mode with limit of {limit} data rows')
        return self.apply(overwrite_derived_data, set_branch_dict, limit, apply_changes=False)

    def explicit_inputs_of_method(self):
        params = {'method_id': self.name}
        wh = ("" if self.batch else "WHERE m.id = $method_id")
        term_clause = """
            OPTIONAL MATCH (class)-[:HAS_CONTROLLED_TERM]->(term:Term),
            (term)<-[:ON_VALUE]-(:Method)<-[:NEXT]-(action)
            """
        q = f"""
        MATCH (m:Method)
        {wh}
        WITH *
        MATCH p=(m)-[:METHOD_ACTION*1..50]->(action:Method) //TODO: check that deep trees >10 never exist 
        WHERE action.type = 'get_data'
        OPTIONAL MATCH (action)-[r:SOURCE_CLASS]->(class:Class)
        {term_clause}
        WITH m, action, collect(CASE WHEN term IS NULL THEN class else term END) as coll1
        OPTIONAL MATCH (action)-[r2:SOURCE_RELATIONSHIP]->(rel:Relationship)-[:TO|:FROM]->(class)
        {term_clause}
        WITH m, coll1, collect(CASE WHEN term IS NULL THEN rel else term END) as coll2
        WITH m, coll1+coll2 as coll
        UNWIND coll as input
        MERGE (m)-[:METHOD_INPUT]->(input)
        """
        self.interface.query(q, params)

    def explicit_outputs_of_method(self):
        params = {'method_id': self.name}
        wh = ("" if self.batch else "WHERE m.id = $method_id")
        q = f"""
        MATCH (m:Method)
        {wh}   
        WITH *
        MATCH p=(m)-[:METHOD_ACTION*1..10]->(action:Method) //TODO: check that deep trees >10 never exist 
        WHERE action.type in ['assign_class']
          OR
          (action.type in ['link'] and (
              EXISTS ( (action)-[:TO_VALUE]->() )
                OR
              EXISTS ( (action)-[:FROM_VALUE]->() )
          ))
          OR
          (action.type in ['link'] and
            EXISTS ( (action)-[:LINK]->() )
          )
        OPTIONAL MATCH (action)-[dir_value]->(term:Term)
        ,(term)<-[:HAS_CONTROLLED_TERM]-(class:Class)
        WHERE type(dir_value) in ["TO_VALUE", "FROM_VALUE"]
        OPTIONAL MATCH (action)-[:LINK]->(rel1:Relationship)-[dir_class]->(class)                
        WHERE type(dir_class) = apoc.text.split(type(dir_value), "_")[0]
        OPTIONAL MATCH (action)-[assign_class:CLASS]->(a_class:Class)
        OPTIONAL MATCH (action)-[rl:LINK]->(rel2:Relationship)
        OPTIONAL MATCH (rel2)-[r2class:TO|FROM]->(class2)
        WHERE type(r2class) = CASE WHEN rl.how in ['merge_from', 'create_from', 'merge_from_on_uri'] THEN 'FROM' ELSE 'TO' END
        WITH DISTINCT 
        m, 
        a_class, 
        CASE WHEN term IS NULL THEN class else term END as output_1,
        rel2 as output_2
        WITH m, collect(output_1) + collect(output_2) + collect(distinct a_class) as coll
        UNWIND coll as output       
        MERGE (m)-[:METHOD_OUTPUT]->(output)               
        """
        self.interface.query(q, params)

    def fetch_metadata(self):
        pass

    def _get_derivation_node_id(self):
        raise NotImplementedError

    def merge_action_json(self, name, action_json_list: list, derivation_method_json: dict = None):
        """
        Merge actions from action_json_list into provided derivation_method_json OR create a new derivation_method_json if not provided.
        :param name: Str name of derivation
        :param action_json_list: List of jsons representing actions
        :param derivation_method_json: json representing a derivation method
        """

        def get_last_action_node_id(method_json):
            """Use topological sorting to find the last action node id from the NEXT rels of the method json"""
            next_rels = []
            for rel in method_json['relationships']:
                if rel.get('type') == 'NEXT':
                    next_rels.append((rel.get('fromId'), rel.get('toId')))
            action_node_ids, _ = topological_sort(next_rels)
            return action_node_ids[-1]
        

        def get_unique_key(node, label, prop_filter):
            if label == 'Method':
                key = f'{node[prop_filter]} {label}'
            else:
                key = f'{node["properties"][prop_filter]} {label}'
            return key


        def rel_key(id_, type_):
            return f'{id_}{type_}'


        if derivation_method_json is None:
            derivation_method_json = {
                "nodes":
                    [
                        {
                            "id": "core0",
                            "position": {},
                            "caption": "",
                            "labels": ["Method"],
                            "properties": {"id": name}
                        }
                    ],
                "relationships": [],
                'style': {}
            }
            previous_method_node_id = None
        else:
            previous_method_node_id = get_last_action_node_id(derivation_method_json)

        filter_node_dict = {
            'Class': 'label',
            'Term': 'Term Code',
            'Method': 'id',
            'Relationship': 'unique_key'  # get set as a temp property below
        }

        # set up some tracking vars for merging nodes
        node_map = {}  # {node_id: node, ...}
        rel_map = {}  # {fromId + type: rel, ...}
        current_nodes = {}  # {'Analysis Value Class': 'n95'} ie {property: id}
        
        # compile a map of node ids
        for node in derivation_method_json['nodes']:
            node_map[node.get('id')] = node

        for rel in derivation_method_json['relationships']:
            # compile a map of relationships
            _key = rel_key(id_=rel.get('fromId'), type_=rel.get('type'))
            rel_map[_key] = rel

        for action_json in action_json_list:
            
            node_indexes_to_skip = []

            for node in action_json['nodes']:
                node_map[node.get('id')] = node

            for rel in action_json['relationships']:
                # compile a map of relationships
                _key = rel_key(id_=rel.get('fromId'), type_=rel.get('type'))
                rel_map[_key] = rel

            for node_index, node in enumerate(action_json['nodes']):
                if 'Method' in node['labels']:

                    # With branch_combine, can get duplicate methods. Skip over them here as they get added to
                    # current_nodes further down
                    if get_unique_key(node, 'Method', 'id') not in current_nodes.keys():

                        # add the METHOD_ACTION relationship from the core/super method node
                        derivation_method_json['relationships'].append(
                            {
                                'id': f'ma_rel_{node["id"]}',
                                'fromId': self._get_derivation_node_id(),
                                'toId': node['id'],
                                'type': 'METHOD_ACTION',
                                'properties': {},
                                'style': {}
                            }
                        )

                        # add the NEXT relationship if this is not the first method node
                        if previous_method_node_id is not None:
                            # check to make sure there isn't already a next from previous json creation.
                            # E.g. get_data and filter
                            if not any(rel for rel in action_json['relationships']
                                        if
                                        (rel.get('fromId') == previous_method_node_id) and (rel.get('type') == 'NEXT')):

                                derivation_method_json['relationships'].append(
                                    {
                                        'id': f'next_rel_{node["id"]}',
                                        'fromId': previous_method_node_id,
                                        'toId': node['id'],
                                        'type': 'NEXT',
                                        'properties': {},
                                        'style': {}
                                    }
                                )

                        previous_method_node_id = node['id']

                # need to add a property 'unique_key' to each rel node in order to test for uniqueness later.
                if 'Relationship' in node.get('labels'):
                    # fetching the node TO and FROM the relationship node to use their labels for a unique key
                    from_key = rel_key(id_=node.get('id'), type_='FROM')
                    to_key = rel_key(id_=node.get('id'), type_='TO')
                    from_rel = rel_map.get(from_key)
                    to_rel = rel_map.get(to_key)
                    from_node = node_map.get(from_rel.get('toId'))
                    to_node = node_map.get(to_rel.get('toId'))

                    # 'from_node_label + rel_node_type + to_node_label' as unique key
                    _key = f'{from_node.get("properties").get("label")}' \
                        f'{node.get("properties").get("relationship_type")}' \
                        f'{to_node.get("properties").get("label")}'

                    node['properties']['unique_key'] = _key

                for label in filter_node_dict.keys():
                    if label in node['labels']:
                        prop_filter = filter_node_dict[label]  # label, relationship_type
                        unique_node_key = get_unique_key(node, label, prop_filter)
                        # eg Analysis Value Class, Analysis Value Relationship
                        # need to check for nodes that could already be present and 'merge' them
                        if unique_node_key in current_nodes.keys():
                            # find all rels to/from this node, then point them to the preexisting node
                            for rel in action_json['relationships']:
                                if rel['fromId'] == node['id']:
                                    # setting the id in the rel to the preexisting node
                                    rel['fromId'] = current_nodes[unique_node_key]
                                elif rel['toId'] == node['id']:
                                    # setting the id in the rel to the preexisting node
                                    rel['toId'] = current_nodes[unique_node_key]
                            # remove that node from the json
                            node_indexes_to_skip.append(node_index)

                        else:  # otherwise add that node to the current nodes to check against later
                            current_nodes[unique_node_key] = node['id']

                # remove the temp unique key added above
                if 'Relationship' in node.get('labels'):
                    del node['properties']['unique_key']

            # add all the nodes and relationships to the derivation method
            derivation_method_json['nodes'].extend([node for i, node in enumerate(action_json['nodes'])
                                                    if i not in node_indexes_to_skip])
            derivation_method_json['relationships'].extend(action_json['relationships'])

            # some cases of an empty rel, ie {}, being in the relationships list. Remove them here. Also remove duplicates.
            unique_rels = []
            unique_rel_keys = set()
            for rel in derivation_method_json['relationships']:
                if rel:
                    _key = rel.get('fromId') + rel.get('type') + rel.get(
                        'toId')  # expecting only one unique rel between two nodes
                    if _key not in unique_rel_keys:
                        unique_rel_keys.add(_key)
                        unique_rels.append(rel)
            derivation_method_json['relationships'] = unique_rels

        return derivation_method_json

    def build_derivation_method_json(self, name=None, actions=None, json_str=True):
        if not name:
            name = self.name
        if not actions:
            self._actions = None
            actions = self.actions

        logger.info(f"Compiling json: {name}")

        action_json_list = [action.retrieve_json() for action in actions]
        derivation_method_json = self.merge_action_json(name, action_json_list)
        if json_str:
            return json.dumps(visualise_json(derivation_method_json), indent=4)
        else:
            return derivation_method_json


class OnlineDerivationMethod(DerivationMethod):
    def __init__(self, name, interface=None, study=None, schema_available=None):
        super().__init__(data=name, name=name, interface=interface, study=study, schema_available=schema_available)
        self.interface.create_index(label='Method', key='parent_id')
        self.batch = self.name == "online"
        self.available_methods = self._available_methods()
        self.applied_methods = self._applied_methods()

    @property
    def actions(self):
        if not self.exists_in_db:
            logger.error(f"Method {self.name} does not exist in database for study {self.study}. "
                         f"Unable to retrieve Actions. First load the method")
        if self._actions is None:
            self._actions = self.retrieve_actions()
        return self._actions

    def _delete_orphaned_nodes(self, detached_nodes_dicts: list):
        node_ids = list(set(reduce(add, [detached_node["detached_nodes"] for detached_node in detached_nodes_dicts])))

        q = f"""
        MATCH (n)
        WHERE id(n) in $node_ids AND NOT n:Term
        OPTIONAL MATCH (n)-[r2]-()
        WHERE NOT type(r2) = "IS_A"
        WITH size(collect(type(r2))) as rel_count, n as n, id(n) as node_id, labels(n) as labels
        WHERE rel_count = 0
        DETACH DELETE n
        RETURN distinct node_id, labels
        """
        params = {"node_ids": node_ids}
        deleted_nodes = self.interface.query(q, params)
        for labels in set([tuple(node["labels"]) for node in deleted_nodes]):
            logger.debug(
                f"\t\tDeleted {len([node for node in deleted_nodes if labels == tuple(node['labels'])])} with Label {labels}")

        q2 = f"""
        MATCH (n)
        WHERE id(n) in $node_ids
        RETURN id(n) as node_id, labels(n) as labels
        """
        remaining_nodes = self.interface.query(q2, params)
        for labels in set([tuple(node["labels"]) for node in remaining_nodes]):
            logger.debug(
                f"\t\tNOT Deleted {len([node for node in remaining_nodes if labels == tuple(node['labels'])])} with label {labels}")

        if len(remaining_nodes) + len(deleted_nodes) != len(node_ids):
            error_msg = f"Number of remaining nodes ({len(remaining_nodes)} + {len(deleted_nodes)} is not equal to {len(node_ids)})"
            logger.error(error_msg)

    def _available_methods(self) -> list:
        q = f"""
            MATCH (m:Method {{parent_id:$study_id}})
            WHERE m.type IS NULL AND NOT (m)<-[:METHOD_ACTION]-() 
            RETURN m.id as method_id
            """
        params = {'study_id': self.study}
        res = self.interface.query(q, params)
        if not res:
            logger.warning(f"There are no methods in database for study {self.study}.")
        return [method.get("method_id") for method in res]

    def _applied_methods(self) -> list:
        q = f"""
            MATCH (m:Method {{parent_id:$study_id}})-[r:METHOD_ACTION]->(action:Method), (action)-[:APPLIED]->(chg:Changes)
            RETURN distinct m.id as method_id
            """
        params = {'study_id': self.study}
        res = self.interface.query(q, params)
        # if not res:
        #     logger.warning(f"There are no applied methods in database for study {self.study}.")
        #     logger.warning(f"The following methods have been loaded but not applied:")
        #     logger.warning(f"{self.available_methods}")
        return [method.get("method_id") for method in res]

    def _get_derivation_node_id(self):
        """All core id nodes in OnlineDerivationMethod will be `core0`"""
        return 'core0'

    def rollback(self):
        if self.applied:
            detached_nodes_dct = []
            # TODO: check if action.applied_metadata_node_id is None for relevant actions
            logger.info(f"{self.name}: Starting Rollback")
            logger.info(f"\tTotal # of Actions: {len(self.actions)}")
            for i, action in enumerate(self.actions):
                logger.info(f"Rolling back Action: #{i + 1} of {len(self.actions)}")
                detached_nodes_dct = action.rollback(detached_nodes_dct)
            if len(detached_nodes_dct) > 0:
                self._delete_orphaned_nodes(detached_nodes_dct)
            logger.info(f"\t\t Completed Rollback!")
        else:
            logger.error(f"Method {self.name} has not been applied for study {self.study}. Nothing to rollback.")

    def delete(self):
        logger.info(f"""Deleting: {self.name if not self.batch else 'ALL derivation methods'}  from the database""")
        if self.batch:
            for method in self.applied_methods:
                OnlineDerivationMethod(name=method, interface=self.interface).rollback()
        elif self.name in self.applied_methods:
            self.rollback()
        params = {"method_id": self.name, "study_id": self.study}
        wh = ("" if self.batch else "WHERE m.id = $method_id and m.parent_id = $study_id")
        actions = [
            f"""
            MATCH (m:Method)
            {wh}
            OPTIONAL MATCH path = (m)-[r:METHOD_ACTION|NEXT*1..10]->(x:`Method`)
            OPTIONAL MATCH path2 = (x)-[r2]->(y)
            //OPTIONAL MATCH (m)-[r*1..10]->(x:`Method`)-[r2]->(y)     
            DETACH DELETE r2, x
            """,
            f"""
            MATCH (m:Method)
            {wh}
            OPTIONAL MATCH (m)-[r]-()
            DELETE r
            DELETE m
            """
        ]
        for q in actions:
            self.interface.query(q, params)

    def explicit_prerequisites_of_method(self):
        """
        Identifies dependenices between Methods and explicitly stores as METHOD_PREREQ relationship
        Caveat:
            only works for -[:METHOD_INPUT]->(r:Relationship)
            for derived Classes we don't have -[:METHOD_INPUT]->(:Class) and should avoid in the future
        :return:
        """
        params = {} if self.batch else {'method_id': self.name}
        wh = "WHERE m.type IS NULL AND NOT (m)<-[:METHOD_ACTION]-()"
        wh += ("" if self.batch else "AND m.id = $method_id")
        q = f"""
        MATCH (m:Method)
        {wh} 
        MATCH (m)-[:METHOD_INPUT]->(r:Relationship)-[:FROM|TO]->(c:Class)
        WHERE NOT (m)-[:METHOD_INPUT]->(c)
        OPTIONAL MATCH (c)-[:HAS_CONTROLLED_TERM]->(t:Term)<-[:METHOD_INPUT]-(m)
        WHERE EXISTS ( ()-[:METHOD_OUTPUT]->(t) )
        WITH m, coalesce(t, r) as input
        MATCH (prereq:Method)
        MATCH (prereq:Method)-[:METHOD_OUTPUT]->()<-[:FROM|TO*0..1]-(input)
        WHERE (prereq.type IS NULL AND NOT (prereq)<-[:METHOD_ACTION]-())
        WITH DISTINCT m, collect(DISTINCT prereq) as coll
        UNWIND coll as prereq
        MERGE (m)-[:METHOD_PREREQ]->(prereq)
        RETURN * 
        """
        self.interface.query(q, params)

    def resolve_methods_order(self, error_cutoff: int = 1000):
        """
        :param error_cutoff:
        :return:
        """

        def get_batch(resolved: list = None):
            if not resolved:
                resolved = []
            q = """
            MATCH (m:Method)
            WHERE m.type IS NULL AND NOT (m)<-[:METHOD_ACTION]-() AND NOT m.id in $resolved
            OPTIONAL MATCH (m)-[:METHOD_PREREQ]->(prereq)
            WHERE NOT prereq.id in $resolved AND NOT prereq = m
            WITH m, collect(prereq) as coll
            WHERE size(coll) = 0
            RETURN m.id as method, id(m) as db_id
            """
            params = {'resolved': resolved}
            res = self.interface.query(q, params)
            return [x['method'] for x in res]

        resolved = []
        p_resolved = None
        i = 0
        while resolved != p_resolved and i < error_cutoff:
            p_resolved = resolved.copy()
            batch = get_batch(resolved=resolved)
            resolved += batch
            i += 1

        return resolved


class DictDerivationMethod(DerivationMethod):
    """
    A DerivationMethod that has been loaded from json external to the database.

    :param content: Dictionary of json representing the DerivationMethod. Contains two keys 'nodes' and 'relationships'
    where each value is a list of dictionaries representing nodes and relationships respectively.
    :param name: The name of the DerivationMethod
    :param interface: A neointerface object for interfacing with the database
    :param study: The name of the study this DerivationMethod is a part of
    :param overwrite_db: Bool representing the decision of overwriting a derivation of the same name in the database on load
    :param check_schema: Bool representing the decision to check on load if a derivation's schema is present in the database
    """


    def __init__(self, content, name=None, interface=None, study=None, overwrite_db=False,
                 schema_available=None, check_schema:bool=False):
        super().__init__(data=content,
                         name=self.get_derivation_name(content) if name is None else name,
                         interface=interface,
                         study=study,
                         schema_available=schema_available)
        
        self.predicted_output_class_info = None
        self.content = content
        self.load(overwrite_db=overwrite_db, check_schema=check_schema)

    @property
    def actions(self):
        if self._actions is None:
            self._actions = self.retrieve_actions()
        return self._actions

    def get_derivation_name(self, content):
        next_rels = reduce(add, [[rel["fromId"], rel["toId"]] for rel in content["relationships"] if
                                 rel["type"] == "NEXT"])

        for node in content["nodes"]:
            if "Method" in node["labels"] and list(node["properties"].keys()) == ["id"] and node["id"] not in next_rels:
                self.name = node["properties"]["id"]
                return self.name
        logger.warning(f"Cannot identify derivation method name from JSON dictionary.")
        raise KeyError

    def _get_derivation_node_id(self):
        next_rels = reduce(add, [[rel["fromId"], rel["toId"]] for rel in self.content["relationships"] if
                                 rel["type"] == "NEXT"], [])

        for node in self.content["nodes"]:
            if "Method" in node["labels"] and list(node["properties"].keys()) == ["id"] and node["id"] not in next_rels:
                return node["id"]
        logger.warning(f"Cannot identify derivation method id from JSON dictionary.")
        raise KeyError

    def load(self, overwrite_db:bool, check_schema: bool = False):
        if self.load_decision(overwrite_db):
            if check_schema:
                self.validate_method_schema()
            if self.exists_in_db:
                OnlineDerivationMethod(name=self.name, interface=self.interface, study=self.study).delete()
            self._load()

    def _load(self):
        super().load()
        self.enrich_existing_relationship_nodes()
        if self.schema_available:
            self.validate_method_dict(self.content["nodes"], self.content["relationships"])
        self.content, merge_on = ModelManager.arrows_dict_uri_dict_enrich(
            dct=simplify_arrows_json(self.content),
            uri_map=ModelManager.URI_MAP)
        self.interface.load_arrows_dict(self.content, merge_on, always_create=['Method'])
        self.post_load_enrichment()

        self._actions = None  # reset action metadata in memory. Next time self.actions is called, it will fetch metadata from database.

        return self.content

    def get_action_type_ids(self, type: str) -> List[int]:
        # Assuming that all links have id of format 'linkx' where x is an integer
        id_integers = [0]
        for node in self.content['nodes']:
            if "properties" in node:
                if node['properties'].get('type') == type:
                    id_int = node['properties']['id'].replace(type, '')
                    id_integers.append(id_int)
        return id_integers

    @property
    def predict_output_classes(self) -> dict:
        if self.predicted_output_class_info is None:
            self.predicted_output_class_info = self._predict_output_classes()
        return self.predicted_output_class_info

    def _predict_output_classes(self) -> dict:
        """
        Predict the output classes of current derivation method (from schema/run_script/assign_class info)
        Runs a limited self.apply() in order to get the most up to date changes node for any run_script actions.
        :return: dict of format {'ASSIGN_CLASSES': [], 'CLASSES AFTER RUNSCRIPT': ['AAGE', 'USUBJID'], 'NEW RUNSCRIPT CLASSES': ['AGEGR'], 'PREDICTED CLASSES': ['AGEGR']}
        """

        self.apply_limited(overwrite_derived_data=False)

        logger.info(f'\tPredicting output classes {self.name}')

        filepath = os.path.dirname(__file__)
        with open(os.path.join(filepath, 'predict_output_classes.cql')) as cypherfile:
            query = cypherfile.read()
        params = {
            "method_id": self.name,
            "study_id": self.study
        }
        res = self.interface.query(query, params)[0]
        
        for key in ['ASSIGN_CLASSES', 'CLASSES AFTER RUNSCRIPT', 'NEW RUNSCRIPT CLASSES', 'PREDICTED CLASSES']:
            if res[key] is None:
                res[key] = []
        
        if not res['PREDICTED CLASSES']:
            logger.debug(f'Query \n {query}')
            logger.debug(f':params {params}')
            logger.warn(f'Predicted 0 output classes for {self.name}')
        else:
            logger.info(f'Predicted classes: {res["PREDICTED CLASSES"]}')
            logger.debug(f'predict_output_classes result: {res}')

        return res

    def _predict_links(self) -> tuple:
        # generate info about predicted links for this derivation method
        logger.info(f'\tPredicting links actions for {self.name}')

        filepath = os.path.dirname(__file__)
        with open(os.path.join(filepath, 'predict_links.cql')) as cypherfile:
            query = cypherfile.read()
        params = self.predict_output_classes

        res = self.interface.query(query, params)
        if not res:
            logger.debug(f'Query \n {query}')
            logger.debug(f'Params {params}')
            logger.warn(f'Predicted 0 links for {self.name}')
            return []
        else:
            res = res[0]['PREDICTED_LINKS']

        new_links = []
        subject_level_classes = []

        for predicted_link in res:
            cls = predicted_link.get('C_PRIMARY')
            from_classes = predicted_link.get('PREDICTED_FROM_CLASSES', [])
            to_classes = predicted_link.get('PREDICTED_TO_CLASSES', [])

            logger.debug(f'\t\tPrimary Class: {cls}')
            logger.debug(f'\t\tPredicted from classes: {from_classes}')
            logger.debug(f'\t\tPredicted from classes: {to_classes}')

            if cls.get('aval_repr', False):
                logger.debug(f'\t\tPrimary Class: {cls} is a Subject Level Class')
                subject_level_classes.append(cls)

            for from_class in from_classes:
                new_links.append(
                    {
                        "from_class": from_class.get('label'),
                        "from_short_label": from_class.get('short_label'),
                        "to_class": cls.get('label'),
                        "to_short_label": cls.get('short_label'),
                        "relationship_type": cls.get('label')  # todo this should be rel type from schema
                    }
                )
            for to_class in to_classes:
                new_links.append(
                    {
                        "from_class": cls.get('label'),
                        "from_short_label": cls.get('short_label'),
                        "to_class": to_class.get('label'),
                        "to_short_label": to_class.get('short_label'),
                        "relationship_type": to_class.get('label')  # todo this should be rel type from schema
                    }
                )

        return new_links, subject_level_classes

    def _generate_link_actions(self) -> list:
        """
        Create Link action objects and then convert them to json.
        """
        logger.info(f'Generating link actions for {self.name}')

        base_link_id = max(self.get_action_type_ids(type='link'))

        predicted_links, subject_level_classes = self._predict_links()

        if not predicted_links:
            return []
        new_link_jsons = []

        for i, link in enumerate(predicted_links, start=1):
            new_link_jsons.append(
                Link(
                    action_dict={
                        "id": f"link{base_link_id+i}",
                        "type": "link",
                        "relationship_type": link.get('relationship_type'),
                        "from_class": link.get('from_class'),
                        "from_short_label": link.get('from_short_label'),
                        "to_class": link.get('to_class'),
                        "to_short_label": link.get('to_short_label'),
                        "to_class_property": 'rdfs:label',
                        "link_how": 'merge'  # link.get('link_how')
                    },
                    interface=self.interface,
                    dont_fetch=True
                ).retrieve_json()
            )

        for i, cls in enumerate(subject_level_classes, start=1):
            new_link_jsons.append(
                SubjectLevelLinkSuperMethod(
                    method=self,
                    interface=self.interface,
                    previous_actions=self.actions,
                    action_info={"type": 'subject_level_link', "id": f"subject_level_link{i}"},
                    action_dict={
                        "m": {
                            "type": 'subject_level_link',
                            "id": f"subject_level_link{i}"
                        },
                        "class": cls,
                        "term": None
                    }
                ).retrieve_json()
            )

        return new_link_jsons

    def merge_predicted_links(self):
        """
        Generate predicted links and merge them with the current method json held in self.content
        Will run a limited (10 row) self.apply() to generate changes nodes for any run_script actions.
        :param load_to_db: Bool that decided if new method json should be loaded into the database
        """

        link_json_list = self._generate_link_actions()
        if link_json_list:
            self.content = self.merge_action_json(
                name=self.name, action_json_list=link_json_list, derivation_method_json=self.content)
            self.load(overwrite_db=True)
            self._actions = None  # so a future call to self.actions will pull new actions from database
        else:
            logger.warn(f'Could not predict links for {self.name}')

    def fetch_get_data_classes(self) -> List[str]:
        """
        Return a list of classes found in GetData actions of this derivation method
        :return: list of class strings
        """
        get_data_classes = []
        for action in self.actions:
            if isinstance(action, GetData):
                get_data_classes += action.meta["source_classes"]
                get_data_classes += [rel.get('from') for rel in action.meta["source_rels"]]
                get_data_classes += [rel.get('to') for rel in action.meta["source_rels"]]
        return get_data_classes

    def _fetch_uri_meta_from_db(self) -> dict:
        predicted_output_classes = self.predict_output_classes
        output_classes = predicted_output_classes.get('PREDICTED CLASSES')  # list of short labels

        q = """
        MATCH (c:Class) 
        WHERE (c.short_label in $output_classes) AND (c.classes_for_uri IS NOT NULL)
        WITH c, [x in split(c.classes_for_uri, "|") | trim(x)] as short_cls_uri
        CALL apoc.cypher.run(
            "MATCH (c_uri:Class) WHERE c_uri.short_label in $short_cls_uri RETURN c_uri",
            {short_cls_uri: short_cls_uri}
        ) YIELD value
        WITH c, apoc.map.fromPairs([['label', value.c_uri.label], ['short_label', value.c_uri.short_label]]) as uri_map
        RETURN c as class, collect(uri_map) as uri_labels
        """
        params = {"output_classes": output_classes}
        res = self.interface.query(q, params)
        # list of class dictionaries e.g. 
        # [
        #   {
        #       'class': {'short_label': 'AAGE', 'label': 'Analysis Age', 'uri': 'neo4j://graph.schema#Class/Analysis+Age', 'derived': 'true', 'classes_for_uri': 'USUBJID|AVISIT'},
        #      'uri_labels': [{'label': 'Subject', 'short_label': 'USUBJID'}, {}, ...]
        #   },
        #   {}, ...
        # ]
        if not res:
            logger.warn(f'Could not find uri metadata for classes: {output_classes}')
        return res

    def _create_build_uri_actions(self) -> list:
        """
        Build a list of json dictionaries that represent build_uri actions based on `classes_for_uri` property on schema classes
        """
        uri_meta = self._fetch_uri_meta_from_db()
        base_uri_id = max(self.get_action_type_ids(type='build_uri'))
        build_uri_actions = []
        for i, item in enumerate(uri_meta, start=1):
            cls = item.get('class')
            uri_labels = [uri.get('label') for uri in item.get('uri_labels')]
            uri_short_labels = [uri.get('short_label') for uri in item.get('uri_labels')]

            build_uri_actions.append(
                BuildUri(
                    action_dict={
                        "id": f"build_uri{base_uri_id+i}",
                        "type": "build_uri",
                        "prefix": "",
                        "prefix": "",
                        "uri_fors": [cls.get("short_label")],
                        "uri_fors_long": [cls.get("label")],
                        "uri_bys": uri_short_labels,
                        "uri_bys_long": uri_labels,
                        "uri_labels": [],
                        "uri_labels_long": []
                    },
                    interface=self.interface,
                    dont_fetch=True
                ).retrieve_json()
            )
        return build_uri_actions

    def merge_build_uri_from_schema(self):
        """
        Add BuildUri actions to the derivation method for classes that have property `classes_for_uri` in the schema.
        Expects `classes_for_uri` property to have string of pipe `|` separated values.
        Will run a limited (10 row) self.apply() to generate changes nodes for any run_script actions.
        :param load_to_db: Bool that decided if new method json should be loaded into the database
        """

        build_uri_list = self._create_build_uri_actions()
        if build_uri_list:
            self.content = self.merge_action_json(
                name=self.name, action_json_list=build_uri_list, derivation_method_json=self.content)
            self.load(overwrite_db=True)
            # ToDo if in the future we run all merge new action functions in one go, 
            # we should only load to db after all predict functions have been run.
        else:
            logger.warn(f'Could not create any BuildUri actions for {self.name}')

    def validate_method_schema(self, method_json: dict = None):
        """
        Raise a KeyError if any schema present in the method_json is not present in the attached Neo4j database.
        :param method_json: json dictionary representing a derivation method. If None then take self.content as method_json.
        """

        logger.info(f'Checking schema is present in database')

        if method_json is None:
            method_json = self.content

        node_dct = {}
        # set up lookup dict for relationship node rels and term rels
        rel_dct_from = {}  # {fromId: {FROM: rel, TO: rel}, ...}  TO and FROM rels that come from a rel_node
        rel_dct_to = {}  # {toId: rel, ...}  rel coming from a class to a term node

        for node in method_json.get('nodes'):
            node_dct[node.get('id')] = node

        for rel in method_json.get('relationships'):
                # set up some relationships maps for later
                if rel.get('type') in ['FROM', 'TO']:
                    # no rel_node should have two FROM (or TO) rels, so unique keys
                    if rel.get('fromId') not in rel_dct_from:
                        rel_dct_from[rel.get('fromId')] = {rel.get('type'): rel}
                    else:
                        rel_dct_from[rel.get('fromId')][rel.get('type')] = rel
                elif rel.get('type') == 'HAS_CONTROLLED_TERM':
                    # no two classes should have the same term, so unique keys
                    rel_dct_to[rel.get('toId')] = rel

        missing_schema_nodes = self._check_method_schema(method_json, node_dct, rel_dct_from, rel_dct_to)

        if missing_schema_nodes.get('Classes') or missing_schema_nodes.get('Terms') or missing_schema_nodes.get(
                'Relationships') or missing_schema_nodes.get('Term Relationships'):
            missing_classes = missing_schema_nodes.get('Classes') if missing_schema_nodes.get('Classes') else ''
            missing_terms = missing_schema_nodes.get('Terms') if missing_schema_nodes.get('Terms') else ''
            missing_rels = missing_schema_nodes.get('Relationships') if missing_schema_nodes.get(
                'Relationships') else ''
            missing_term_rels = missing_schema_nodes.get('Term Relationships') if missing_schema_nodes.get(
                'Term Relationships') else ''
            logger.error(f'Some of the schema required for {self.name} could not be found in the connected database!\n {missing_classes=}\n {missing_rels=}\n {missing_terms=}\n {missing_term_rels=}')
            raise KeyError

    
    def _check_method_schema(self, json, node_dct, rel_dct_from, rel_dct_to) -> dict:
        """
        Return a dictionary of any schema from the derivation method that is not present in the attached Neo4j database.
        """
        
        missing_nodes = {
            'Terms': [],
            'Classes': [],
            'Relationships': [],
            'Term Relationships': []
        }
        decode_node_dct = {}
        decode_class_dct = {}

        for node in json.get('nodes'):

            if 'decode' == node.get('properties').get('type'):
                decode_node_dct[node.get('id')] = node

            elif 'Class' in node.get('labels'):
                missing_nodes['Classes'].append(
                    {
                        'label': node.get('properties').get('label')
                        }
                    )

            elif 'Term' in node.get('labels'):
                has_controlled_term_rel = rel_dct_to[node.get('id')]
                term_class_id = has_controlled_term_rel.get('fromId')
                class_label = node_dct[term_class_id].get('properties').get('label')

                term_dct = {
                    'Class Label': class_label,
                    'Codelist Code': node.get('properties').get('Codelist Code'),
                    'Term Code': node.get('properties').get('Term Code')
                }
                if node.get('labels'):
                    if 'Study Specific Term' not in node.get('labels'):  # allow non-existence of Study-Specific Terms
                        missing_nodes['Terms'].append(term_dct)

            elif 'Relationship' in node.get('labels'):
                from_rel_class_id = rel_dct_from[node.get('id')].get('FROM').get('toId')
                to_rel_class_id = rel_dct_from[node.get('id')].get('TO').get('toId')

                from_class_label = node_dct[from_rel_class_id].get('properties').get('label')
                to_class_label = node_dct[to_rel_class_id].get('properties').get('label')

                rel_type = node.get('properties').get('relationship_type')
                rel_dct = {'from': from_class_label, 'to': to_class_label, 'relationship_type': rel_type}

                missing_nodes['Relationships'].append(rel_dct)

        for rel in json.get('relationships'):
            if rel.get('fromId') in decode_node_dct:
                try:
                    decode_class_dct[rel.get('fromId')]
                except KeyError:
                    decode_class_dct[rel.get('fromId')] = {}
                # there is a decode method, so get its from and to class labels
                if rel.get('type') == 'FROM_CLASS':
                    decode_from_class_label = node_dct.get(rel.get('toId')).get('properties').get('label')
                    decode_class_dct[rel.get('fromId')]['FROM_CLASS'] = decode_from_class_label
                if rel.get('type') == 'TO_CLASS':
                    decode_to_class_label = node_dct.get(rel.get('toId')).get('properties').get('label')
                    decode_class_dct[rel.get('fromId')]['TO_CLASS'] = decode_to_class_label

        if missing_nodes['Classes']:
            # check to see if classes are in the database.
            missing_nodes['Classes'] = self._check_classes_against_db(missing_nodes['Classes'])

        if missing_nodes['Relationships']:
            # check to see if relationships are in the database.
            missing_nodes['Relationships'] = self._check_rels_against_db(missing_nodes['Relationships'])

        if missing_nodes['Terms']:
            # check to see if terms are in the database.
            missing_nodes['Terms'] = self._check_terms_against_db(missing_nodes['Terms'])

        if decode_class_dct:
            # check to see if SAME_AS rels exist between these classes
            missing_nodes['Term Relationships'] = self._check_term_schema_against_db(decode_class_dct)

        return missing_nodes

    def _check_term_schema_against_db(self, decode_class_dct):

        missing_class_pairs = []
        for _, classes in decode_class_dct.items():
            q = '''
                MATCH (from_class:Class), (to_class:Class)
                WHERE from_class.label = $FROM_CLASS AND to_class.label = $TO_CLASS
                MATCH (from_class)-[:HAS_CONTROLLED_TERM]->(from_class_term:Term)
                ,(to_class)-[:HAS_CONTROLLED_TERM]->(to_class_term:Term)
                ,(from_class_term)-[r_sa:SAME_AS]->(to_class_term)
                RETURN r_sa
                '''
            params = classes
            res = self.interface.query(q, params)
            if not res:
                missing_class_pairs.append([classes.get('FROM_CLASS'), classes.get('TO_CLASS')])
            return missing_class_pairs

    def _check_rels_against_db(self, rel_dcts):
        # check to see if a list of relationship exist in the database
        missing_rels = []
        for rel in rel_dcts:
            q = '''
            MATCH path = (c_from:Class)<-[:FROM]-(rel:Relationship)-[:TO]->(c_to:Class)
            WHERE c_from.label = $from AND c_to.label = $to AND rel.relationship_type = $relationship_type
            RETURN path
            '''
            params = rel
            res = self.interface.query(q, params)
            if not res:
                missing_rels.append(rel)
        return missing_rels

    def _check_classes_against_db(self, class_dicts):
        # check to see if a list of classes exist in the database
        missing_classes = []
        for cls in class_dicts:
            q = '''
            MATCH (c:Class)
            WHERE c.label = $label
            RETURN c
            '''
            params = cls
            res = self.interface.query(q, params)
            if not res:
                missing_classes.append(cls)
        return missing_classes

    def _check_terms_against_db(self, term_dcts):
        # check to see if a list of terms exist in the database
        missing_terms = []
        for term in term_dcts:
            q = '''
            MATCH path = (c:Class)-[:HAS_CONTROLLED_TERM]->(t:Term)
            WHERE c.label = $`Class Label` AND t.`Term Code` = $`Term Code`
            RETURN path
            '''
            params = term
            res = self.interface.query(q, params)
            if not res:
                missing_terms.append(term)
        return missing_terms

    def validate_method_dict(self, nodes, relationships):

        assert "nodes" in self.content.keys(), f"{self.name}: {self.content}"
        assert "relationships" in self.content.keys(), f"{self.name}: {self.content}"

        method_ids = []
        methods_by_type = {}
        rels = {}
        classes = {}
        terms = {}

        # collecting nodes in dicts
        for node in nodes:
            assert "labels" in node.keys(), f"{self.name}: {node}"
            assert "properties" in node.keys(), f"{self.name}: {node}"
            assert "id" in node.keys(), f"{self.name}: {node}"
            if "Method" in node["labels"]:
                assert "id" in node["properties"].keys(), f"Method does not have an id: {self.name}: {node}"
                method_ids.append(node["properties"]['id'])
                if "type" in node["properties"]:
                    node_type = node["properties"]["type"]
                else:
                    node_type = ""
                if node_type not in methods_by_type.keys():
                    methods_by_type[node_type] = []
                methods_by_type[node_type].append({
                    "node": node,
                    "fromRels": self._validate_method_dict_get_rels(node["id"], "fromId", relationships),
                    "toRels": self._validate_method_dict_get_rels(node["id"], "toId", relationships)
                })
            if "Relationship" in node["labels"]:
                rels[node["id"]] = {
                    "node": node,
                    "fromRels": self._validate_method_dict_get_rels(node["id"], "fromId", relationships),
                    "toRels": self._validate_method_dict_get_rels(node["id"], "toId", relationships)
                }
            if "Class" in node["labels"]:
                classes[node["id"]] = {
                    "node": node,
                    "fromRels": self._validate_method_dict_get_rels(node["id"], "fromId", relationships),
                    "toRels": self._validate_method_dict_get_rels(node["id"], "toId", relationships)
                }
            if "Term" in node["labels"]:
                terms[node["id"]] = {
                    "node": node,
                    "fromRels": self._validate_method_dict_get_rels(node["id"], "fromId", relationships),
                    "toRels": self._validate_method_dict_get_rels(node["id"], "toId", relationships)
                }

        assert "" in methods_by_type.keys(), f"{self.name} Method must have a core Method node with no type assigned, {self.content}"

        method_ids_seen = set()
        assert len(method_ids) == len(set(method_ids)), \
            f"Found duplicate method ids: {[x for x in method_ids if x in method_ids_seen or method_ids_seen.add(x)]}"

        core_method_found = False
        for m in methods_by_type[""]:
            if m["node"]["properties"]["id"] == self.name:
                core_method_found = True
            for r in m["fromRels"] + m["toRels"]:
                assert r["type"] in ["NEXT",
                                     "METHOD_ACTION"], f"{self.name} Unexpected relationship {r} with Method {m}"
        assert core_method_found, f"Core method (with id {self.name}) not found"
        get_data_actions = ["get_data"]
        if not any([m in methods_by_type.keys() for m in get_data_actions]):
            logging.warning(
                f"{self.name} Method should have at least one get_data_action {get_data_actions}: {self.content}")
        else:
            for m in methods_by_type["get_data"]:
                for r in m["fromRels"]:
                    assert r["type"] in ["NEXT", "SOURCE_CLASS", "SOURCE_RELATIONSHIP"], \
                        f"{self.name} Unexpected relationship {r} from Method {m}"
                    if r["type"] == "SOURCE_RELATIONSHIP":
                        assert r["toId"] in rels.keys(), \
                            f"{self.name} Node with label Relationship is expected at SOURCE_RELATIONSHIP: {r}"
                    if r["type"] == "SOURCE_CLASS":
                        assert r["toId"] in classes.keys(), \
                            f"{self.name} Node with label Class is expected at SOURCE_CLASS: {r}"
                for r in m["toRels"]:
                    assert r["type"] in ["NEXT",
                                         "METHOD_ACTION"], f"{self.name} Unexpected relationship {r} to Method {m}"

        filter_ms = methods_by_type.get("filter")
        if not filter_ms:
            filter_ms = []
        for m in filter_ms:
            for r in m["fromRels"]:
                assert r["type"] in ["NEXT", "ON", "ON_VALUE", "FILTER_RELATIONSHIP", "ONLY_RELATED_TO"], \
                    f"{self.name} Unexpected relationship {r} from Method {m}"
                if r["type"] == "ON":
                    assert r["toId"] in classes.keys(), \
                        f"{self.name} Node with label Class is expected at ON: {r}"
                if r["type"] == "ON_VALUE":
                    assert r["toId"] in terms.keys(), \
                        f"{self.name} Node with label Term is expected at ON_VALUE: {r}"
            for r in m["toRels"]:
                assert r["type"] in ["NEXT",
                                     "METHOD_ACTION"], f"{self.name} Unexpected relationship {r} to Method {m}"

        write_data_actions = ["link", "assign_class", "apply_stat", "decode", "subject_level_link"]
        if not any([m in methods_by_type.keys() for m in write_data_actions]):
            logging.warning(
                f"{self.name} Method should have at least one write_data_action {write_data_actions}, it had: {self._fetch_action_names_from_content()}")
        else:
            link_ms = methods_by_type.get("link")
            if not link_ms:
                link_ms = []
            for m in link_ms:
                for r in m["fromRels"]:
                    assert r["type"] in ["NEXT", "LINK", "TO_VALUE", "FROM_VALUE"], \
                        f"{self.name} Unexpected relationship {r} from Method {m}"
                    if r["type"] == "LINK":
                        assert r["toId"] in rels.keys(), \
                            f"{self.name} Node with label Relationship is expected at LINK: {r}"
                    if r["type"] == "TO_VALUE":
                        assert r["toId"] in terms.keys(), \
                            f"{self.name} Node with label Term is expected at TO_VALUE: {r}"
                    if r["type"] == "FROM_VALUE":
                        assert r["toId"] in terms.keys(), \
                            f"{self.name} Node with label Term is expected at FROM_VALUE: {r}"
                for r in m["toRels"]:
                    assert r["type"] in ["NEXT",
                                         "METHOD_ACTION"], f"{self.name} Unexpected relationship {r} to Method {m}"
            ac_ms = methods_by_type.get("assign_class")
            if not ac_ms:
                ac_ms = []
            for m in ac_ms:
                for r in m["fromRels"]:
                    assert r["type"] in ["NEXT", "ON", "CLASS"], \
                        f"{self.name} Unexpected relationship {r} from Method {m}"
                    if r["type"] in ["ON", "CLASS"]:
                        assert r["toId"] in classes.keys(), \
                            f"{self.name} Node with label Class is expected on ON/CLASS: {r}"
                for r in m["toRels"]:
                    assert r["type"] in ["NEXT",
                                         "METHOD_ACTION"], f"{self.name} Unexpected relationship {r} to Method {m}"

        # Validating Relationships
        for rel in rels.values():
            rel_TO_rels = [x for x in rel["fromRels"] if x["type"] == "TO"]
            rel_FROM_rels = [x for x in rel["fromRels"] if x["type"] == "FROM"]
            assert len(
                rel_TO_rels) == 1, f"{self.name} Relationship must have exactly 1 :TO class: {rel}: {rel_TO_rels}"
            assert len(
                rel_FROM_rels) == 1, f"{self.name} Relationship must have exactly 1 :FROM class: {rel}: {rel_TO_rels}"
            assert rel_TO_rels[0]["toId"] in classes.keys(), \
                f"{self.name} Relationship must TO-point to a Class: {rel}: {rel_TO_rels[0]}"
            assert rel_FROM_rels[0]["toId"] in classes.keys(), \
                f"{self.name} Relationship must FROM-point to a Class: {rel}: {rel_FROM_rels[0]}"

        # # Validating apply_stat
        # if "apply_stat" in methods_by_type.keys():
        #     for apply_stat_method in methods_by_type['apply_stat']:
        #         as_rels_by_type = {}
        #         for rel in rels:
        #             if rel["fromId"] == apply_stat_method['node']['id']:
        #                 if rel["type"] not in as_rels_by_type.keys():
        #                     as_rels_by_type[rel["type"]] = []
        #                 as_rels_by_type[rel["type"]].append(rel)
        #         assert as_rels_by_type.get("Result"), \
        #             f"{id} apply_stat Method must specify at least 1 Result: {apply_stat_method}"
        #         assert as_rels_by_type.get("Statistic"), \
        #             f"{id} apply_stat Method must specify at least 1 Statistic: {apply_stat_method}"

        return True

    @staticmethod
    def _validate_method_dict_get_rels(nodeid: str, key: str, rels: dict):
        res = []
        for r in rels:
            if r.get(key) == nodeid:
                res.append(r)
        return res

    def _fetch_action_names_from_content(self):
        """Go through self.content and retrieve action types for methods (actions)"""
        return [node['properties'].get('type') for node in self.content.get('nodes', list()) if ('Method' in node.get('labels')) and (node['properties'].get('type', False))]


class SnippetDerivationMethod(DictDerivationMethod):
    """
    Subclass of DictDerivationMethod which instead of getting passed json directly, takes in a list of actions and
    subsequently generates its own json.
    """
    def __init__(self, name: str, actions: list, interface: NeoInterface = None, study=None, overwrite_db=False,
                 schema_available=None):
        content = self.build_derivation_method_json(name, actions, json_str=False)
        super().__init__(content=content,
                         name=name,
                         interface=interface,
                         study=study,
                         overwrite_db=overwrite_db,
                         schema_available=schema_available)


class RdfDerivationMethod(DerivationMethod):
    def __init__(self, content, name=None, interface=None, study=None, overwrite_db=False, schema_available=None):
        super().__init__(data=content, name=name, interface=interface, study=study, schema_available=schema_available)
        self.content = content
        if self.load_decision(overwrite_db):
            if self.exists_in_db:
                OnlineDerivationMethod(name=self.name, interface=self.interface, study=self.study).delete()
            self.load()

    @property
    def actions(self):
        if self._actions is None:
            self._actions = self.retrieve_actions()
        return self._actions

    def load(self):
        super().load()
        self.interface.rdf_generate_uri(ModelManager.URI_MAP)
        res = self.interface.rdf_import_subgraph_inline(self.content)
        self.post_load_enrichment()
        return res



import collections


def compare_unordered_lists(l1: [], l2: []) -> bool:
    """
    Compare two lists regardless of order of elements.
    Duplicates elements, if present, are treated as completely separate.

    IMPORTANT:  the elements of the list must match exactly,
                and must be HASHABLE Python entities, such as
                strings, numbers or tuples; they can NOT be, for example, dictionaries

    Return True if the given lists match, as defined above; False, otherwise.

    EXAMPLES:   [1, 2, 3] will match [3, 2, 1]
                [] and [] will match
                ["x", (1, 2)] will match [(1, 2) , "x"] but NOT ["x", (2, 1)]
                ["a", "a"] will NOT match ["a"]

    :param l1:  A list of HASHABLE Python entities (e.g. strings or numbers)
    :param l2:  Same as above
    :return:    True if there's a match, or False otherwise
    """
    return collections.Counter(l1) == collections.Counter(l2)



def compare_recordsets(rs1: [{}], rs2: [{}]) -> bool:
    """
    We define "recordsets" as "lists of dictionaries".  Each element of the lists is regarded as a "record".

    EXAMPLE of recordset:  [{'Field_A': 1},
                            {'Field_A': 1},
                            {'Field_A': 99, 'Field_B': 'hello'}]

    Compare 2 recordsets WITHOUT REGARD to the position of the dictionaries within the lists,
    and also WITHOUT REGARD to the position of the key:value pairs within the dictionaries.
    Duplicates records, if present, are treated as completely separate.

    Return True if the given recordsets match, as defined above; False, otherwise.

    WARNING: this function is meant for comparing SMALL datasets, because it's Order n square!

    :param rs1: A (possibly empty) list of dictionaries
    :param rs2: A (possibly empty) list of dictionaries

    :return:    True if there's a match, or False otherwise
    """

    # Verify the type of the arguments
    assert isinstance(rs1, list), "compare_recordsets() : The 1st argument is not a list!  Value = " + str(rs1)
    assert isinstance(rs2, list), "compare_recordsets() : The 2nd argument is not a list!  Value = " + str(rs2)

    if len(rs1) != len(rs2):
        return False    # Datasets of different sizes will never match

    # Consider each element (i.e. a dictionary) in turn in the first list:
    #   attempt to remove it from the other list; if the removal fails, then it means
    #   that we have an element in the 1st list that is not present in the 2nd one (hence a mismatch)
    for rec1 in rs1:
        # Note: since Python 3.7 dictionaries are order-preserving, but
        #       built-in Python functions such as "remove"
        #       do not distinguish dictionaries based on order:
        #           {'a': 1, 'b': 2} will match {'b': 2, 'a': 1}

        try:
            rs2.remove(rec1)    # Remove (the first instance of) the element rec1 from the list rs2
        except Exception:
            return False        # The remove failed - i.e. the first list contains an element not in the 2nd one

    return True


def format_nodes(method_json: dict, keep_id=False):
    keys_to_remove =  ['style', 'position', 'caption'] if keep_id else  ['id', 'style', 'position', 'caption']
    nodes = []
    for i in method_json['nodes']:
        temp = {}
        for k, v in i.items():
            if (k not in keys_to_remove) and v:  # also remove empty entries e.g. properties: {}
                temp[k] = v
        nodes.append(temp)
        del temp
    return nodes


def format_relationships(method_json: dict, keep_id=False):
    keys_to_remove =  ['style', 'position', 'caption'] if keep_id else  ['id', 'style', 'position', 'caption', 'fromId', 'toId']
    relationships = []
    for i in method_json['relationships']:
        temp = {}
        for k, v in i.items():
            if (k not in keys_to_remove) and v:  # also remove empty entries e.g. properties: {}
                temp[k] = v
        relationships.append(temp)
        del temp
    return relationships


def format_json(method_json: dict, keep_id=False) -> dict:
    method_json['nodes'] = format_nodes(method_json, keep_id=keep_id)
    method_json['relationships'] = format_relationships(method_json, keep_id=keep_id)
    if 'style' in method_json:
        del method_json['style']
    return method_json


def compare_method_json(method_json, compare_method_json):
    # remove ids, style and captions from json
    method_json = format_json(method_json)
    compare_method_json = format_json(compare_method_json)

    nodes = method_json['nodes']
    compare_nodes = compare_method_json['nodes']
    rels = method_json['relationships']
    compare_rels = compare_method_json['relationships']

    # Datasets of different sizes will never match
    assert len(nodes) == len(compare_nodes)
    assert len(rels) == len(compare_rels)

    # sort keys of node and rel dicts
    sorted_nodes = [{key: sorted(value.items()) if isinstance(value, dict) else value for key, value in sorted(node.items())} for node in nodes]
    sorted_rels = [{key: sorted(value.items()) if isinstance(value, dict) else value for key, value in sorted(rel.items())} for rel in rels]
    sorted_compare_nodes = [{key: sorted(value.items()) if isinstance(value, dict) else value for key, value in sorted(node.items())} for node in compare_nodes]
    sorted_compare_rels = [{key: sorted(value.items()) if isinstance(value, dict) else value for key, value in sorted(rel.items())} for rel in compare_rels]

    # Consider each element (i.e. a dictionary) in turn in the first list:
    #   attempt to remove it from the other list; if the removal fails, then it means
    #   that we have an element in the 1st list that is not present in the 2nd one (hence a mismatch)
    for node in sorted_nodes:
        try:
            sorted_compare_nodes.remove(node)
        except Exception:
            raise KeyError(f'Could not remove node from json: {node} \n\n{sorted_nodes=} \n{sorted_compare_nodes=} \n\n{sorted_rels=} \n{sorted_compare_rels=}') # The remove failed - i.e. there is a difference in the node lists

    for rel in sorted_rels:
        try:
            sorted_compare_rels.remove(rel)
        except Exception:
            raise KeyError(f'Could not remove rel from json: {rel} \n\n{sorted_nodes=} \n{sorted_compare_nodes=} \n\n{sorted_rels=} \n{sorted_compare_rels=}') # The remove failed - i.e. there is a difference in the node lists

    return True



##########################  TESTS  ##########################


def test_compare_unordered_lists():
    # Tests for the compare_unordered_lists function

    # POSITIVE tests
    assert compare_unordered_lists([1, 2, 3] , [1, 2, 3])
    assert compare_unordered_lists([1, 2, 3] , [3, 2, 1])
    assert compare_unordered_lists([] , [])
    assert compare_unordered_lists( ["x", (1, 2)]  ,  [(1, 2) , "x"] )

    # NEGATIVE tests
    assert not compare_unordered_lists( ["x", (1, 2)]  ,  ["x", (2, 1)] )
    assert not compare_unordered_lists( ["a", "a"]  ,  ["a"] )




def test_compare_recordsets():
    # Tests for the compare_recordsets function

    # POSITIVE tests

    assert compare_recordsets(  [  {'patient_id': 444, 'gender': 'F', 'condition_id': 'happy'} ,  {'gender': 'M' , 'patient_id': 123} ],
                                [  {'patient_id': 444, 'gender': 'F', 'condition_id': 'happy'} ,  {'gender': 'M' , 'patient_id': 123} ]
                             )  # Everything absolutely identical

    assert compare_recordsets(  [  {'patient_id': 444, 'gender': 'F', 'condition_id': 'happy'} ,  {'gender': 'M' , 'patient_id': 123} ],
                                [  {'patient_id': 444, 'gender': 'F', 'condition_id': 'happy'} ,  {'patient_id': 123 , 'gender': 'M'} ]
                             )  # 2 fields reversed in last record of 2nd dataset

    assert compare_recordsets(  [  {'patient_id': 444, 'gender': 'F', 'condition_id': 'happy'} ,  {'gender': 'M' , 'patient_id': 123} ],
                                [   {'gender': 'M' , 'patient_id': 123} , {'patient_id': 444, 'gender': 'F', 'condition_id': 'happy'} ]
                             )  # Records reversed in 2nd dataset

    assert compare_recordsets(  [  {'patient_id': 444, 'gender': 'F', 'condition_id': 'happy'} ,  {'gender': 'M' , 'patient_id': 123} ],
                                [   {'patient_id': 123 , 'gender': 'M'} , {'patient_id': 444, 'gender': 'F', 'condition_id': 'happy'}  ]
                             )  # Records reversed in 2nd dataset, and fields reversed in one of them

    assert compare_recordsets(  [  {'gender': 'F', 'patient_id': 444, 'condition_id': 'happy'} ,  {'gender': 'M' , 'patient_id': 123} ],
                                [   {'patient_id': 123 , 'gender': 'M'} , {'patient_id': 444, 'condition_id': 'happy', 'gender': 'F'}  ]
                             )  # Additional order scrambling in the last test

    assert compare_recordsets([] , [])      # 2 empty datasets

    assert compare_recordsets([{'a': 1}] , [{'a': 1}])      # Minimalist data sets!

    assert compare_recordsets([{'a': 1}, {'a': 1}]  ,
                              [{'a': 1}, {'a': 1}])      # Each dataset has 2 identical records

    assert compare_recordsets([{'a': 1}, {'a': 1}, {'z': 'hello'}]  ,
                              [{'z': 'hello'}, {'a': 1}, {'a': 1}])     # Scrambled record order, with duplicates


    # NEGATIVE tests

    assert not compare_recordsets(  [  {'patient_id': 444, 'gender': 'F', 'condition_id': 'happy'} ,  {'gender': 'M' , 'patient_id': 123} ],
                                    [  {'patient_id': 444, 'gender': 'F', 'condition_id': 'happy'}  ]
                                 )  # Missing record in 2nd dataset

    assert not compare_recordsets(  [  {'patient_id': 444, 'gender': 'F', 'condition_id': 'happy'} ,  {'gender': 'M' , 'patient_id': 123} ],
                                    [  {'patient_id': 444, 'gender': 'F', 'condition_id': 'happy'} ,  {'patient_id': 123} ]
                                 ) # Missing field in last record of 2nd dataset

    assert not compare_recordsets(  [  {'patient_id': 444, 'gender': 'F', 'condition_id': 'happy'} ,  {'gender': 'M' , 'patient_id': 123} ],
                                    [  {'patient_id': 444, 'gender': 'F', 'condition_id': 'happy'} ,  {'gender': 'M' , 'patient_id': 123, 'extra_field': 'some junk'} ]
                                 ) # Extra field in 2nd dataset

    assert not compare_recordsets(  [  {'patient_id': 444, 'gender': 'F', 'condition_id': 'happy'} ,  {'gender': 'M' , 'patient_id': 123} ],
                                    [  {'patient_id': 444, 'gender': 'F', 'condition_id': 'happy'} ,  {'gender': 'M' , 'patient_id': 123} ,  {'extra_record': 'what am I doing here?'} ]
                                 )  # Extra record in 2nd dataset

    assert not compare_recordsets( [] , [{'a': 1}] )      # one empty dataset and one non-empty

    assert not compare_recordsets([{'a': 1}]  ,
                                  [{'a': 1}, {'a': 1}])    # 1 record is NOT the same things as 2 identical ones

    assert not compare_recordsets([{'a': 1}, {'a': 1}, {'z': 'hello'}]  ,
                                  [{'a': 1}, {'a': 1}])     # datasets of different size



from abc import abstractmethod

from derivation_method.action import GetData, Link, RunScript, AssignLabel, CallAPI, BuildUri, BranchLoad, \
    BranchSave, BranchCombine, LinkStat, RunCypher


class Method:
    interface = NotImplemented

    def __init__(self):
        self.branch_dfs = None

    def retrieve_actions(self):
        actions = []
        q = """
        MATCH path=(m:Method{id:$method_id, parent_id:$parent_id})-[:METHOD_ACTION]->(action:`Method`),
        actions=(action)-[:NEXT*0..99]->(last_action:`Method`)
        WHERE NOT ( (last_action)-[:NEXT]->() ) and NOT ( ()-[:NEXT]->(action) )
        RETURN [nd in nodes(actions) | {node_id: id(nd), type: nd.type, id: nd.id}] as map
        """
        params = self.build_retrieve_actions_params()
        res = self.interface.query(q, params)
        if res:
            action_dicts = res[0]["map"]
        else:
            action_dicts = []
        for i, action_dict in enumerate(action_dicts):
            if action_dict["type"] == "filter":
                assert isinstance(actions[-1], GetData), "Filter actions can only come after GetData actions"
            elif action_dict["type"] == "get_data":
                if action_dicts[i + 1]["type"] == "filter":
                    filter_dict = action_dicts[i + 1]
                    actions.append(GetData(action_dict, filter_dict, self))
                else:
                    actions.append(GetData(action_dict, None, self))
            elif action_dict["type"] == "run_script":
                actions.append(RunScript(action_dict, self))
            elif action_dict["type"] == "call_api":
                actions.append(CallAPI(action_dict, self))
            elif action_dict["type"] == "assign_class":
                actions.append(AssignLabel(action_dict, self))
            elif action_dict["type"] == "link":
                actions.append(Link(action_dict, self))
            elif action_dict["type"] == "run_cypher":
                actions.append(RunCypher(action_dict, self))
            elif action_dict["type"] == "build_uri":
                actions.append(BuildUri(action_dict, self))
            elif action_dict["type"] == "branch_save":
                actions.append(BranchSave(action_dict, self))
            elif action_dict["type"] == "branch_load":
                actions.append(BranchLoad(action_dict, self))
            elif action_dict["type"] == "branch_combine":
                actions.append(BranchCombine(action_dict, self))
            elif action_dict["type"] == "link_stat":
                actions.append(LinkStat(action_dict, self))
            elif action_dict["type"] == "apply_stat":
                apply_stat = self.handle_supermethod(action_dict, previous_actions=actions)
                actions.append(apply_stat)
                # actions.extend(apply_stat.actions)
            elif action_dict["type"] == "decode":
                decode = self.handle_supermethod(action_dict, previous_actions=actions)
                actions.append(decode)
            elif action_dict["type"] == "subject_level_link":
                subject_level_link = self.handle_supermethod(action_dict, previous_actions=actions)
                actions.append(subject_level_link)
            elif action_dict["type"] is None:
                super_method_actions = self.handle_supermethod(action_dict)
                actions.append(super_method_actions)
            else:
                raise ValueError(f"Unknown action type {action_dict['type']}")
        return actions

    @abstractmethod
    def handle_supermethod(self, action_dict, previous_actions=None):
        pass

    @abstractmethod
    def fetch_metadata(self):
        pass

    @abstractmethod
    def build_retrieve_actions_params(self):
        pass



import re
from neointerface.neointerface import NeoInterface
import networkx as nx
from collections import defaultdict


def id_to_integer(node_id):
    try:
        return int("".join(char for char in node_id if char.isdigit()))
    except ValueError:
        raise ValueError(f"invalid ID: {node_id}")


def simplify_arrows_json(raw_json):
    to_remove = ["style", "position", "caption"]
    new_json = {}
    for key, value in raw_json.items():
        if key == "style":
            continue
        elif key in ["nodes", "relationships"]:
            new_json[key] = sorted([{
                kkey: vvalue
                for kkey, vvalue in item.items() if kkey not in to_remove
            } for item in value], key=lambda x: x["id"])
        else:
            new_json[key] = value
    return new_json


def visualise_json(method_json, vis_type: str = 'rigid'):

    def kamada_kawai_layout(_json):
        nx_graph = nx.Graph()
        for node in _json['nodes']:
            nx_graph.add_node(node.get('id'))
        for rel in _json['relationships']:
            nx_graph.add_edge(rel.get('fromId'), rel.get('toId'))
        return nx.kamada_kawai_layout(nx_graph)

    vis_types = ['fluid', 'rigid']
    assert vis_type in vis_types, \
        f'vis_type of {vis_type} was not a valid vis_type! Expected value in {vis_types}'
    m_count, c_count, r_count, t_count, o_count = 0, 0, 0, 0, 0

    if vis_type == 'fluid':
        base_distance = max(500, len(method_json['nodes'])*10)
        position_dict = kamada_kawai_layout(method_json)
        for node in method_json['nodes']:
            node_id = node.get('id')
            x = position_dict.get(node_id)[0]*base_distance
            y = position_dict.get(node_id)[1]*base_distance
            node['position'] = {'x': x, 'y': y}
    else:
        for count, node in enumerate(method_json['nodes']):
            base_distance = 300
            if 'core' in node['id']:
                x = 0
                y = 0
            elif 'Method' in node['labels']:
                x = m_count * base_distance
                m_count += 1
                y = base_distance
            elif 'Relationship' in node['labels']:
                x = r_count * base_distance
                r_count += 1
                y = base_distance * 2
            elif 'Class' in node['labels']:
                x = c_count * base_distance
                c_count += 1
                y = base_distance * 3
            elif 'Term' in node['labels']:
                x = t_count * base_distance
                t_count += 1
                y = base_distance * 4
            else:
                x = o_count * base_distance
                o_count += 1
                y = base_distance * 5

            node['position'] = {'x': x, 'y': y}

    for node in method_json['nodes']:
        if 'core' in node['id']:
            colour = '#68bc00'  # green
        elif 'Method' in node['labels']:
            colour = '#9f0500'  # red
        elif 'Relationship' in node['labels']:
            colour = '#009ce0'  # blue
        elif 'Class' in node['labels']:
            colour = '#000000'  # black
        elif 'Term' in node['labels']:
            colour = '#666666'  # silver
        else:
            colour = '#fa28ff'  # pink
        node['style'] = {'border-color': colour}

    for relationship in method_json['relationships']:
        if relationship:  # is not empty
            if relationship['type'] == 'METHOD_ACTION':
                colour = '#68bc00'
            elif relationship['type'] == 'NEXT':
                colour = '#9f0500'
            elif relationship['type'] == 'SOURCE_RELATIONSHIP':
                colour = '#009ce0'
            elif relationship['type'] in ['HAS_CONTROLLED_TERM', 'ON_VALUE', 'TO_VALUE', 'FROM_VALUE']:
                colour = '#666666'
            elif relationship['type'] in ['FILTER_RELATIONSHIP']:
                colour = '#dbdf00'
            else:
                colour = '#000000'
            relationship['style'] = {'arrow-color': colour}
    return method_json


def get_arrows_json_cypher(
        neo: NeoInterface,
        q: str,
        params: dict = None,
        keep_labels: list = None,
        hide_labels: list = None,
        keep_props: dict = None,
        hide_props: dict = None,
        x_shift=0,
        y_shift=0,
        limit=None,
        step=500
):
    assert limit is None or isinstance(limit, int)
    assert isinstance(step, int)
    assert (keep_labels is None) != (hide_labels is None), 'cannot use keep_labels and hide_labels together, use one or the other'
    assert (keep_props is None) != (hide_props is None), 'cannot use keep_props and hide_props together, use one or the other'
    assert step > 0
    if not params:
        params = {}
    if not hide_labels:
        hide_labels = []
    if not hide_props:
        hide_props = {}
    filter_labels = keep_labels if keep_labels else hide_labels
    filter_props = keep_props if keep_props else hide_props
    params = {**params, **{'filter_labels': filter_labels, 'filter_props': filter_props, 'x_shift': x_shift, 'y_shift': y_shift}}
    q = f"""
        //apoc.any.properties() around virtual nodes is req to get properties
        //apoc.node.labels() to get labels of virtual nodes
        //apoc.rel.startNode/endNode for virt relationships
        {q}
        WITH *
        ORDER BY id(x), id(y), id(r) 
        WITH 
            apoc.coll.toSet(collect(distinct x) + collect(distinct y)) as nds, 
            collect(distinct r) as rls  
        WITH 
        [x in nds | 
            {{		
                id: "n" + toString(id(x)),
                labels: [y in apoc.node.labels(x) WHERE {'' if keep_labels else 'NOT'} y in $filter_labels], //all but the ones in hide_labels
                properties: //all but the ones in hide_props (for each label) {{<label>: LIST OF <prop>}}
                    apoc.map.submap(
                        apoc.any.properties(x), 
                        [key in keys(apoc.any.properties(x)) WHERE {'' if keep_props else 'NOT'} key in 
                            (apoc.coll.toSet(
                                apoc.coll.flatten(
                                    [key in apoc.coll.intersection(apoc.node.labels(x), keys($filter_props)) | $filter_props[key]]
                                )
                            ))
                        ]
                    ),
                caption:"",
                style: {{}}
            }}
        ] as nds, 
        [r in [y in rls WHERE NOT y IS NULL] |
            {{
                id: "r" + toString(id(r)),
                type: type(r),
                style: {{}},
                properties: apoc.any.properties(r),
                fromId: "n" + toString(id(apoc.rel.startNode(r))),
                toId: "n" + toString(id(apoc.rel.endNode(r)))
            }}
        ] as rls
        WITH *, size(nds) as sz, 200 as step
        WITH *, toInteger(ceil(sqrt(sz))) as axlen
        WITH *, apoc.coll.flatten([x in range(1, axlen) | [y in range(1, axlen) | {{position: {{x:$x_shift+(-axlen/2+x)*step, y:$y_shift+(-axlen/2+y)*step}}}}]][0..apoc.coll.max([sz-1,1])]) as positions // grid
        //WITH *, [i in range(0, sz-1) | {{position: {{x: $x_shift+round(step*(1+toFloat(i)/10)*cos((2*pi()/10)*i)), y: $y_shift+round(step*(1+toFloat(i)/10)*sin((2*pi()/10)*i))}}}}] as positions // spiral
        WITH *, [pair in apoc.coll.zip(nds,positions) | apoc.map.merge(pair[0],pair[1])] as nds, rls
        RETURN 
            nds as nodes, 
            rls as relationships, 
            {{}} as style
        """
    res = neo.query(q, params)
    return res


def _check_on_keys_dict(node: dict, on_keys_dict: dict):
    return all(node[key] == value for key, value in on_keys_dict.items())


def _update_relationship(relationship_dict: dict, old_id, new_id):
    return {key: value if (key not in ["fromId", "toId"] or value != old_id) else new_id for key, value in
            relationship_dict.items()}


def merge_dicts_on_node_keys(left_dict: dict, right_dict: dict, on_keys_dict: dict):
    try:
        left_dict_common_node = [node for node in left_dict["nodes"] if _check_on_keys_dict(node, on_keys_dict)][0]
        right_dict_common_node = [node for node in right_dict["nodes"] if _check_on_keys_dict(node, on_keys_dict)][0]
    except IndexError:
        raise ValueError("dictionaries didn't satisfy on_keys_dict")

    output_nodes = left_dict["nodes"] + [node for node in right_dict["nodes"] if node != right_dict_common_node]
    output_relationships = left_dict["relationships"] + [
        _update_relationship(rel, right_dict_common_node["id"], left_dict_common_node["id"]) for rel in
        right_dict["relationships"]]

    result = left_dict.copy()
    result.update({"nodes": output_nodes, "relationships": output_relationships})
    return result


# This method checks if not both _id_<name> and <name> column exist in GetData and returns the missing columns
def add_warn_log_if_column_missing(col_names : list):
    missing_cols = set()
    for i in col_names:
        if i.startswith('_id_'):
            if not i[4:] in col_names:
                missing_cols.add(i[4:])
        else:
            if not "_id_"+i in col_names:
                missing_cols.add("_id_"+i)
    return missing_cols


def topological_sort(dependency_pairs):
    """Sort values subject to dependency constraints"""
    num_heads = defaultdict(int)  # num arrows pointing in
    tails = defaultdict(list)  # list of arrows going out
    for h, t in dependency_pairs:
        num_heads[t] += 1
        tails[h].append(t)

    ordered = [h for h in tails if h not in num_heads]
    for h in ordered:
        for t in tails[h]:
            num_heads[t] -= 1
            if not num_heads[t]:
                ordered.append(t)
    cyclic = [n for n, heads in num_heads.items() if heads]
    return ordered, cyclic


def extract_classes_from_query(query: str):
    pattern = r'\((\`?)(?:(?:\w|\s)+)?\1\:(.+?)\)'
    classes = sum(
        [re.sub(
            r'(`)|(\{.+?\}$)',
            '',
            x[1].replace("\n", "").strip()
        ).split(":") for x in re.findall(pattern, query, re.DOTALL)],
        []
    )
    return classes
